from collections import defaultdict
from datetime import timedelta
from itertools import combinations, count
from math import log10
from types import MappingProxyType

from dateutil.relativedelta import relativedelta

from sqlalchemy import Float, cast, or_, select
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import dumps, loads

import chellow.bank_holidays
from chellow.e.computer import hh_rate
from chellow.models import (
    BillType,
    Contract,
    GBill,
    GContract,
    GEra,
    GRateScript,
    GReadType,
    GRegisterRead,
)
from chellow.utils import (
    HH,
    PropDict,
    hh_after,
    hh_before,
    hh_max,
    hh_min,
    hh_range,
    to_ct,
    utc_datetime,
    utc_datetime_now,
)


def get_times(sess, caches, start_date, finish_date, forecast_date):
    times_cache = get_g_engine_cache(caches, "times")
    try:
        s_cache = times_cache[start_date]
    except KeyError:
        s_cache = times_cache[start_date] = {}

    try:
        f_cache = s_cache[finish_date]
    except KeyError:
        f_cache = s_cache[finish_date] = {}

    try:
        return f_cache[forecast_date]
    except KeyError:
        if start_date > finish_date:
            raise BadRequest("The start date is after the finish date.")
        times_dict = defaultdict(int)
        dt = finish_date
        years_back = 0
        while dt > forecast_date:
            dt -= relativedelta(years=1)
            years_back += 1

        times_dict["history-finish"] = dt
        times_dict["history-start"] = dt - (finish_date - start_date)

        times_dict["years-back"] = years_back

        f_cache[forecast_date] = times_dict
        return times_dict


def get_g_engine_cache(caches, name):
    try:
        return caches["g_engine"][name]
    except KeyError:
        caches["g_engine"] = defaultdict(dict)
        return caches["g_engine"][name]


def g_contract_func(caches, contract, func_name):
    try:
        ns = caches["g_engine"]["funcs"][contract.id]
    except KeyError:
        try:
            contr_func_cache = caches["g_engine"]["funcs"]
        except KeyError:
            contr_func_cache = get_g_engine_cache(caches, "funcs")

        try:
            ns = contr_func_cache[contract.id]
        except KeyError:
            ns = {"db_id": contract.id, "properties": contract.make_properties()}
            exec(contract.charge_script, ns)
            contr_func_cache[contract.id] = ns

    return ns.get(func_name, None)


def forecast_date():
    now = utc_datetime_now()
    return utc_datetime(now.year, now.month, 1)


def get_data_sources(ds, start_date, finish_date, forecast_date=None):
    if forecast_date is None:
        forecast_date = ds.forecast_date

    if (
        ds.start_date == start_date
        and ds.finish_date == finish_date
        and forecast_date == ds.forecast_date
    ):
        yield ds
    else:
        for g_era in ds.sess.query(GEra).filter(
            GEra.g_supply == ds.g_supply,
            GEra.start_date <= finish_date,
            or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
        ):
            chunk_start = hh_max(g_era.start_date, start_date)
            chunk_finish = hh_min(g_era.finish_date, finish_date)

            ds = GDataSource(
                ds.sess,
                chunk_start,
                chunk_finish,
                forecast_date,
                g_era,
                ds.caches,
                ds.g_bill,
            )
            yield ds


def datum_range(sess, caches, years_back, start_date, finish_date):
    try:
        return caches["g_engine"]["datum"][years_back][start_date][finish_date]
    except KeyError:
        try:
            g_engine_cache = caches["g_engine"]
        except KeyError:
            g_engine_cache = caches["g_engine"] = {}

        try:
            d_cache_datum = g_engine_cache["datum"]
        except KeyError:
            d_cache_datum = g_engine_cache["datum"] = {}

        try:
            d_cache_years = d_cache_datum[years_back]
        except KeyError:
            d_cache_years = d_cache_datum[years_back] = {}

        try:
            d_cache = d_cache_years[start_date]
        except KeyError:
            d_cache = d_cache_years[start_date] = {}

        datum_list = []
        bank_holidays_id = Contract.get_non_core_by_name(sess, "bank_holidays").id
        for dt in hh_range(caches, start_date, finish_date):
            hist_date = dt - relativedelta(years=years_back)
            ct_dt = to_ct(dt)

            utc_is_month_end = (dt + HH).day == 1 and dt.day != 1
            ct_is_month_end = (ct_dt + HH).day == 1 and ct_dt.day != 1

            utc_decimal_hour = dt.hour + dt.minute / 60
            ct_decimal_hour = ct_dt.hour + ct_dt.minute / 60

            bhs = hh_rate(sess, caches, bank_holidays_id, dt)["bank_holidays"]

            bank_holidays = [b[5:] for b in bhs]
            utc_is_bank_holiday = dt.strftime("%m-%d") in bank_holidays
            ct_is_bank_holiday = ct_dt.strftime("%m-%d") in bank_holidays

            datum_list.append(
                MappingProxyType(
                    {
                        "hist_start": hist_date,
                        "start_date": dt,
                        "ct_day": ct_dt.day,
                        "utc_month": dt.month,
                        "utc_day": dt.day,
                        "utc_decimal_hour": utc_decimal_hour,
                        "utc_year": dt.year,
                        "utc_hour": dt.hour,
                        "utc_minute": dt.minute,
                        "ct_year": ct_dt.year,
                        "ct_month": ct_dt.month,
                        "ct_decimal_hour": ct_decimal_hour,
                        "ct_day_of_week": ct_dt.weekday(),
                        "utc_day_of_week": dt.weekday(),
                        "utc_is_bank_holiday": utc_is_bank_holiday,
                        "ct_is_bank_holiday": ct_is_bank_holiday,
                        "utc_is_month_end": utc_is_month_end,
                        "ct_is_month_end": ct_is_month_end,
                        "status": "X",
                        "kwh": 0,
                        "hist_kwh": 0,
                        "unit_code": "M3",
                        "unit_factor": 1,
                        "units_consumed": 0,
                        "correction_factor": 1,
                        "aq": 0,
                        "soq": 0,
                        "calorific_value": 0,
                        "avg_cv": 0,
                    }
                )
            )
        datum_tuple = tuple(datum_list)
        d_cache[finish_date] = datum_tuple
        return datum_tuple


ACTUAL_READ_TYPES = ["A", "C", "S"]
CORRECTION_FACTOR = 1.02264


def g_industry_rates(sess, caches, g_contract_id_or_name, date):
    return g_rates(sess, caches, g_contract_id_or_name, date, True)


def g_supplier_rates(sess, caches, g_contract_id_or_name, date):
    return g_rates(sess, caches, g_contract_id_or_name, date, False)


def g_rates(sess, caches, g_contract_id_or_name, date, is_industry):
    try:
        return caches["g_engine"]["rates"][g_contract_id_or_name][date]
    except KeyError:
        try:
            ccache = caches["g_engine"]
        except KeyError:
            ccache = caches["g_engine"] = {}

        try:
            rss_cache = ccache["rates"]
        except KeyError:
            rss_cache = ccache["rates"] = {}

        try:
            i_cache = rss_cache[is_industry]
        except KeyError:
            i_cache = rss_cache[is_industry] = {}

        try:
            cont_cache = i_cache[g_contract_id_or_name]
        except KeyError:
            cont_cache = i_cache[g_contract_id_or_name] = {}

        try:
            return cont_cache[date]
        except KeyError:
            if isinstance(g_contract_id_or_name, int):
                if is_industry:
                    g_contract = GContract.get_industry_by_id(
                        sess, g_contract_id_or_name
                    )
                else:
                    g_contract = GContract.get_supplier_by_id(
                        sess, g_contract_id_or_name
                    )
            elif isinstance(g_contract_id_or_name, str):
                if is_industry:
                    g_contract = GContract.get_industry_by_name(
                        sess, g_contract_id_or_name
                    )
                else:
                    g_contract = GContract.get_supplier_by_name(
                        sess, g_contract_id_or_name
                    )
            else:
                raise BadRequest("g_contract_id_or_name must be an int or str")

            month_after = date + relativedelta(months=1) + relativedelta(days=1)
            month_before = date - relativedelta(months=1) - relativedelta(days=1)

            rs = sess.execute(
                select(GRateScript).where(
                    GRateScript.g_contract == g_contract,
                    GRateScript.start_date <= date,
                    or_(
                        GRateScript.finish_date == null(),
                        GRateScript.finish_date >= date,
                    ),
                )
            ).scalar_one_or_none()

            if rs is None:
                rs = (
                    sess.execute(
                        select(GRateScript)
                        .where(GRateScript.g_contract == g_contract)
                        .order_by(GRateScript.start_date.desc())
                    )
                    .scalars()
                    .first()
                )
                if date < rs.start_date:
                    cstart = month_before
                    cfinish = min(month_after, rs.start_date - HH)
                else:
                    cstart = max(rs.finish_date + HH, month_before)
                    cfinish = month_after
            else:
                cstart = max(rs.start_date, month_before)
                if rs.finish_date is None:
                    cfinish = month_after
                else:
                    cfinish = min(rs.finish_date, month_after)

            prefix = "industry" if g_contract.is_industry else "supplier"

            vals = PropDict(
                f"the rate script {chellow.utils.url_root}/g/{prefix}_rate_scripts/"
                f"{rs.id} ",
                loads(rs.script),
                [],
            )
            for dt in hh_range(caches, cstart, cfinish):
                if dt not in cont_cache:
                    cont_cache[dt] = vals

            return vals


def _read_generator(sess, g_supply, start, is_forwards, is_prev):
    if is_prev:
        r_typ = GRegisterRead.prev_type
        r_dt = GRegisterRead.prev_date
    else:
        r_typ = GRegisterRead.pres_type
        r_dt = GRegisterRead.pres_date

    q = (
        sess.query(GRegisterRead)
        .join(GBill)
        .join(BillType)
        .join(r_typ)
        .filter(
            GReadType.code.in_(ACTUAL_READ_TYPES),
            GBill.g_supply == g_supply,
            BillType.code != "W",
        )
    )

    if is_forwards:
        q = q.filter(r_dt >= start).order_by(r_dt, GRegisterRead.id)
    else:
        q = q.filter(r_dt < start).order_by(r_dt.desc(), GRegisterRead.id)

    for offset in count():
        r = q.offset(offset).first()
        if r is None:
            break

        if is_prev:
            dt = r.prev_date
            vl = r.prev_value
        else:
            dt = r.pres_date
            vl = r.pres_value

        g_bill = (
            sess.query(GBill)
            .join(BillType)
            .filter(
                GBill.g_supply == g_supply,
                GBill.g_reads.any(),
                GBill.finish_date >= r.g_bill.start_date,
                GBill.start_date <= r.g_bill.finish_date,
                BillType.code != "W",
            )
            .order_by(GBill.issue_date.desc(), BillType.code, GBill.reference.desc())
            .first()
        )

        if g_bill.id != r.g_bill.id:
            continue

        yield {"date": dt, "value": vl, "msn": r.msn}


class GDataSource:
    def __init__(
        self, sess, start_date, finish_date, forecast_date, g_era, caches, g_bill
    ):
        self.sess = sess
        self.caches = caches
        self.forecast_date = forecast_date
        self.start_date = start_date
        self.finish_date = finish_date
        self.bill_hhs = {}
        times = get_times(sess, caches, start_date, finish_date, forecast_date)
        self.years_back = times["years-back"]
        self.history_start = times["history-start"]
        self.history_finish = times["history-finish"]

        self.problem = ""
        self.bill = defaultdict(int, {"problem": ""})
        self.hh_data = []
        self.rate_sets = defaultdict(set)

        self.g_bill = g_bill
        if self.g_bill is not None:
            self.g_bill_start = g_bill.start_date
            self.g_bill_finish = g_bill.finish_date
            self.is_last_g_bill_gen = (
                not self.g_bill_finish < self.start_date
                and not self.g_bill_finish > self.finish_date
            )

        self.g_era = g_era
        self.g_supply = g_era.g_supply
        self.mprn = self.g_supply.mprn
        self.g_exit_zone_code = self.g_supply.g_exit_zone.code
        self.g_ldz_code = self.g_supply.g_exit_zone.g_ldz.code
        self.g_dn_code = self.g_supply.g_exit_zone.g_ldz.g_dn.code
        self.account = g_era.account
        self.g_reading_frequency = g_era.g_reading_frequency
        self.g_reading_frequency_code = self.g_reading_frequency.code
        self.g_contract = g_era.g_contract

        self.consumption_info = ""

        if self.years_back == 0:
            hist_g_eras = [self.g_era]
        else:
            hist_g_eras = (
                sess.query(GEra)
                .filter(
                    GEra.g_supply == self.g_supply,
                    GEra.start_date <= self.history_finish,
                    or_(
                        GEra.finish_date == null(),
                        GEra.finish_date >= self.history_start,
                    ),
                )
                .order_by(GEra.start_date)
                .all()
            )
            if len(hist_g_eras) == 0:
                hist_g_eras = (
                    sess.query(GEra)
                    .filter(GEra.g_supply == self.g_supply)
                    .order_by(GEra.start_date)
                    .limit(1)
                    .all()
                )

        hist_map = {}

        for i, hist_g_era in enumerate(hist_g_eras):
            if self.history_start > hist_g_era.start_date:
                chunk_start = self.history_start
            else:
                if i == 0:
                    chunk_start = self.history_start
                else:
                    chunk_start = hist_g_era.start_date

            chunk_finish = hh_min(hist_g_era.finish_date, self.history_finish)
            if self.g_bill is None:
                self.consumption_info += _no_bill_kwh(
                    sess,
                    caches,
                    self.g_supply,
                    chunk_start,
                    chunk_finish,
                    hist_g_era,
                    self.g_ldz_code,
                    hist_map,
                    forecast_date,
                )
            else:
                _bill_kwh(
                    sess,
                    self.caches,
                    self.g_supply,
                    hist_g_era,
                    chunk_start,
                    chunk_finish,
                    hist_map,
                    self.g_ldz_code,
                )

        for d in datum_range(
            sess, self.caches, self.years_back, start_date, finish_date
        ):
            h = d.copy()
            hist_start = h["hist_start"]
            h.update(hist_map.get(hist_start, {}))
            h["kwh"] = (
                h["units_consumed"]
                * h["unit_factor"]
                * h["correction_factor"]
                * h["calorific_value"]
                / 3.6
            )
            h["ug_rate"] = float(
                g_rates(sess, self.caches, "ug", h["start_date"], True)[
                    "ug_gbp_per_kwh"
                ][self.g_exit_zone_code]
            )
            self.hh_data.append(h)
            self.bill_hhs[d["start_date"]] = {}

    def g_industry_rates(self, g_contract_id_or_name, date):
        return g_industry_rates(self.sess, self.caches, g_contract_id_or_name, date)

    def g_supplier_rates(self, g_contract_id_or_name, date):
        return g_supplier_rates(self.sess, self.caches, g_contract_id_or_name, date)


def _no_bill_kwh(
    sess,
    caches,
    g_supply,
    start,
    finish,
    g_era,
    g_ldz_code,
    hist_map,
    forecast_date,
):
    read_list = []
    pairs = []
    read_keys = set()

    for is_forwards in (False, True):
        prev_reads = iter(_read_generator(sess, g_supply, start, is_forwards, True))
        pres_reads = iter(_read_generator(sess, g_supply, start, is_forwards, False))
        if is_forwards:
            read_list.reverse()

        for read in _make_reads(is_forwards, prev_reads, pres_reads):
            read_key = read["date"], read["msn"]
            if read_key in read_keys:
                continue
            read_keys.add(read_key)

            read_list.append(read)
            pair = _find_pair(is_forwards, read_list)
            if pair is not None:
                pairs.append(pair)
                if not is_forwards or (is_forwards and read_list[-1]["date"] > finish):
                    break

    consumption_info = "read list - \n" + dumps(read_list) + "\n"
    hhs = _find_hhs(sess, caches, g_era, pairs, start, finish, g_ldz_code)
    _set_status(hhs, read_list, forecast_date)
    hist_map.update(hhs)
    return consumption_info + "pairs - \n" + dumps(pairs)


def _find_pair(is_forwards, read_list):
    if len(read_list) < 2:
        return

    if is_forwards:
        back, front = read_list[-2], read_list[-1]
    else:
        back, front = read_list[-1], read_list[-2]

    back_date, back_value = back["date"], back["value"]
    front_date, front_value = front["date"], front["value"]

    if back["msn"] == front["msn"]:
        units = float(front_value - back_value)
        num_hh = (front_date - back_date).total_seconds() / (30 * 60)

        # Clocked?
        if units < 0:
            digits = int(log10(back_value)) + 1
            units += 10**digits

        return {"start-date": back_date, "units": units / num_hh}


def _bill_kwh(
    sess, caches, g_supply, hist_g_era, chunk_start, chunk_finish, hist_map, g_ldz_code
):
    cf = float(hist_g_era.correction_factor)
    aq = float(hist_g_era.aq)
    soq = float(hist_g_era.soq)
    g_unit = hist_g_era.g_unit
    unit_code, unit_factor = g_unit.code, float(g_unit.factor)

    for hh_date in hh_range(caches, chunk_start, chunk_finish):
        cv, avg_cv = find_cv(sess, caches, hh_date, g_ldz_code)
        hist_map[hh_date] = {
            "unit_code": unit_code,
            "unit_factor": unit_factor,
            "correction_factor": cf,
            "aq": aq,
            "soq": soq,
            "calorific_value": cv,
            "avg_cv": avg_cv,
        }

    g_bills = dict(
        (b.id, b)
        for b in sess.query(GBill)
        .filter(
            GBill.g_supply == g_supply,
            GBill.start_date <= chunk_finish,
            GBill.finish_date >= chunk_start,
        )
        .order_by(GBill.issue_date.desc(), GBill.start_date)
    )
    while True:
        to_del = None
        for a, b in combinations(g_bills.values(), 2):
            if all(
                (
                    a.start_date == b.start_date,
                    a.finish_date == b.finish_date,
                    a.kwh == -1 * b.kwh,
                    a.net == -1 * b.net,
                    a.vat == -1 * b.vat,
                    a.gross == -1 * b.gross,
                )
            ):
                to_del = (a.id, b.id)
                break
        if to_del is None:
            break
        else:
            for k in to_del:
                del g_bills[k]

    for _, g_bill in sorted(g_bills.items()):
        units_consumed = 0
        for prev_value, pres_value in sess.query(
            cast(GRegisterRead.prev_value, Float), cast(GRegisterRead.pres_value, Float)
        ).filter(GRegisterRead.g_bill == g_bill):
            units_diff = pres_value - prev_value
            if units_diff < 0:
                total_units = 10 ** len(str(int(prev_value)))
                c_units = total_units - prev_value + pres_value
                if c_units < abs(units_diff):
                    units_diff = c_units

            units_consumed += units_diff

        bill_s = (
            g_bill.finish_date - g_bill.start_date + timedelta(minutes=30)
        ).total_seconds()
        hh_units_consumed = units_consumed / (bill_s / (60 * 30))

        block_start = hh_max(g_bill.start_date, chunk_start)
        block_finish = hh_min(g_bill.finish_date, chunk_finish)
        for hh_date in hh_range(caches, block_start, block_finish):
            hist_map[hh_date]["units_consumed"] = hh_units_consumed


def find_cv(sess, caches, dt, g_ldz_code):
    cvs = g_rates(sess, caches, "cv", dt, True)["cvs"][g_ldz_code]
    ct = to_ct(dt)
    try:
        cv_props = cvs[ct.day]
    except KeyError:
        cv_props = sorted(cvs.items())[-1][1]

    cv = float(cv_props["cv"])

    try:
        avg_cv = caches["g_engine"]["avg_cvs"][g_ldz_code][ct.year][ct.month]
    except KeyError:
        try:
            gec = caches["g_engine"]
        except KeyError:
            gec = caches["g_engine"] = {}

        try:
            avg_cache = gec["avg_cvs"]
        except KeyError:
            avg_cache = gec["avg_cvs"] = {}

        try:
            avg_cvs_cache = avg_cache[g_ldz_code]
        except KeyError:
            avg_cvs_cache = avg_cache[g_ldz_code] = {}

        try:
            year_cache = avg_cvs_cache[ct.year]
        except KeyError:
            year_cache = avg_cvs_cache[ct.year] = {}

        try:
            avg_cv = year_cache[ct.month]
        except KeyError:
            cv_list = [float(v["cv"]) for v in cvs.values()]
            avg_cv = year_cache[ct.month] = sum(cv_list) / len(cv_list)
    return cv, avg_cv


def _find_hhs(sess, caches, hist_g_era, pairs, chunk_start, chunk_finish, g_ldz_code):
    hhs = {}
    if len(pairs) == 0:
        pairs.append({"start-date": chunk_start, "units": 0})

    # set finish dates
    for i in range(1, len(pairs)):
        pairs[i - 1]["finish-date"] = pairs[i]["start-date"] - HH
    pairs[-1]["finish-date"] = None

    # stretch
    if hh_after(pairs[0]["start-date"], chunk_start):
        pairs[0]["start-date"] = chunk_start

    # chop
    if hh_before(pairs[0]["finish-date"], chunk_start):
        del pairs[0]
    if hh_after(pairs[-1]["start-date"], chunk_finish):
        del pairs[-1]

    # squash
    if hh_before(pairs[0]["start-date"], chunk_start):
        pairs[0]["start-date"] = chunk_start
    if hh_after(pairs[-1]["finish-date"], chunk_finish):
        pairs[-1]["finish-date"] = chunk_finish

    cf = float(hist_g_era.correction_factor)
    aq = float(hist_g_era.aq)
    soq = float(hist_g_era.soq)
    g_unit = hist_g_era.g_unit
    unit_code, unit_factor = g_unit.code, float(g_unit.factor)
    for pair in pairs:
        units = pair["units"]
        for hh_date in hh_range(caches, pair["start-date"], pair["finish-date"]):
            cv, avg_cv = find_cv(sess, caches, hh_date, g_ldz_code)

            hhs[hh_date] = {
                "unit_code": unit_code,
                "unit_factor": unit_factor,
                "units_consumed": units,
                "correction_factor": cf,
                "aq": aq,
                "soq": soq,
                "calorific_value": cv,
                "avg_cv": avg_cv,
            }
    return hhs


def _make_reads(forwards, prev_reads, pres_reads):
    prev_read = next(prev_reads, None)
    pres_read = next(pres_reads, None)
    while prev_read is not None or pres_read is not None:
        if prev_read is None:
            yield pres_read
            pres_read = next(pres_reads, None)

        elif pres_read is None:
            yield prev_read
            prev_read = next(prev_reads, None)

        else:
            if (forwards and prev_read["date"] < pres_read["date"]) or (
                not forwards and prev_read["date"] >= pres_read["date"]
            ):
                yield prev_read
                prev_read = next(prev_reads, None)
            else:
                yield pres_read
                pres_read = next(pres_reads, None)


def _set_status(hhs, read_list, forecast_date):
    THRESHOLD = 31 * 48 * 30 * 60
    rl = [r for r in read_list if r["date"] <= forecast_date]
    for k, v in hhs.items():
        try:
            periods = (abs(r["date"] - k).total_seconds() for r in rl)
            next(p for p in periods if p <= THRESHOLD)
            v["status"] = "A"
        except StopIteration:
            v["status"] = "E"



from collections import defaultdict
from decimal import Decimal
from io import BytesIO

from chellow.e.bill_parsers.sse_edi import Parser, _process_CCD2, _process_MTR
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_missing_mpan_core(mocker, sess):
    file_lines = [
        "MHD=2+UTLBIL:3'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    bills = parser.make_raw_bills()
    expected_bill = {
        "bill_type_code": None,
        "account": None,
        "mpan_core": None,
        "reference": None,
        "issue_date": None,
        "start_date": None,
        "finish_date": None,
        "kwh": Decimal("0"),
        "net": Decimal("0.00"),
        "vat": Decimal("0.00"),
        "gross": Decimal("0.00"),
        "breakdown": {"raw-lines": ["MHD=2+UTLBIL:3'", "MTR=6'"]},
        "reads": [],
    }
    assert bills == [expected_bill]


def test_parser_REF(mocker, sess):
    file_lines = [
        "REF='",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()


def test_process_CCD2(mocker, sess):
    headers = {"reads": [], "breakdown": defaultdict(int), "kwh": Decimal(0)}
    elements = {
        "TMOD": ["0001"],
        "MTNR": ["x"],
        "MLOC": ["228477911812004111222"],
        "PRDT": ["201001"],
        "PVDT": ["200901"],
        "PRRD": ["10", "00", "0", "00"],
        "ADJF": ["", "0"],
        "CONA": ["0"],
        "NUCT": ["0"],
        "CPPU": ["0"],
        "CTOT": ["0"],
    }
    _process_CCD2(elements, headers)
    expected_headers = {
        "reads": [
            {
                "msn": "x",
                "mpan": "04 111 222 22 8477 9118 120",
                "coefficient": Decimal("0"),
                "units": "kWh",
                "tpr_code": "00001",
                "prev_date": utc_datetime(2020, 9, 1, 22, 30),
                "prev_value": Decimal("0"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 10, 1, 22, 30),
                "pres_value": Decimal("10"),
                "pres_type_code": "N",
            }
        ],
        "breakdown": {
            "00001-kwh": Decimal("0"),
            "00001-rate": {Decimal("0")},
            "00001-gbp": Decimal("0"),
        },
        "kwh": Decimal("0"),
    }
    assert headers == expected_headers


def test_process_MTR_ebatch(mocker, sess):
    issue_date = to_utc(ct_datetime(2020, 10, 1))
    start_date = to_utc(ct_datetime(2020, 9, 1))
    finish_date = to_utc(ct_datetime(2020, 9, 30, 23, 30))
    bill_type_code = "N"
    mpan_core = "22 8477 9118 129"
    account = "AC2"
    reference = "h98ge4kl"
    kwh = Decimal(23)
    net = Decimal("45.02")
    vat = Decimal("5.27")
    gross = Decimal("27.64")
    breakdown = {"raw-lines": []}

    headers = {
        "is_ebatch": True,
        "reads": [{"pres_type_code": "C"}],
        "message_type": "UTLBIL",
        "mpan_core": mpan_core,
        "bill_type_code": bill_type_code,
        "account": account,
        "reference": reference,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh,
        "net": net,
        "vat": vat,
        "gross": gross,
        "breakdown": breakdown,
    }
    elements = {}
    bill = _process_MTR(elements, headers)
    expected_bill = {
        "bill_type_code": bill_type_code,
        "account": account,
        "mpan_core": mpan_core,
        "reference": reference,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh,
        "net": net,
        "vat": vat,
        "gross": gross,
        "breakdown": breakdown,
        "reads": [{"pres_type_code": "E"}],
    }
    assert bill == expected_bill



from decimal import Decimal
from io import BytesIO

from werkzeug.exceptions import BadRequest

from chellow.e.bill_parsers.haven_edi import (
    BillElement,
    Parser,
    _process_BTL,
    _process_CCD1,
    _process_CCD3,
    _process_CLO,
    _process_MAN,
    _process_MHD,
    _process_MTR,
    _process_segment,
)
from chellow.utils import utc_datetime


def test_process_BTL_zeroes(mocker):
    elements = {
        "UVLT": ["0"],
        "UTVA": ["0"],
        "TBTL": ["0"],
    }
    headers = {}
    _process_BTL(elements, headers)
    expected_headers = {
        "net": Decimal("0.00"),
        "vat": Decimal("0.00"),
        "gross": Decimal("0.00"),
    }
    assert headers == expected_headers
    assert str(headers["net"]) == "0.00"


def test_process_BTL_non_zeroes(mocker):
    elements = {
        "UVLT": ["11"],
        "UTVA": ["12"],
        "TBTL": ["10"],
    }
    headers = {}
    _process_BTL(elements, headers)
    expected_headers = {
        "net": Decimal("0.11"),
        "vat": Decimal("0.12"),
        "gross": Decimal("0.10"),
    }
    assert headers == expected_headers


def test_process_MTR_UTLHDR(mocker):
    elements = {}
    headers = {"message_type": "UTLHDR", "breakdown": {}}
    bill = _process_MTR(elements, headers)
    assert bill is None


def test_process_MTR_UTLBIL(mocker):
    MockSupply = mocker.patch("chellow.e.bill_parsers.haven_edi.Supply", autospec=True)
    mock_supply = mocker.Mock()
    MockSupply.get_by_mpan_core.return_value = mock_supply
    mock_era = mocker.Mock()
    mock_era.ssc.code = "0244"
    gbp = "10.31"
    cons = "113"
    mock_supply.find_era_at.return_value = mock_era
    elements = {}
    sess = mocker.Mock()
    headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            )
        ],
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "Day",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            },
        ],
        "bill_type_code": "N",
    }
    expected_headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            )
        ],
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {
            "00206-gbp": Decimal(gbp),
            "00206-rate": {Decimal("0.0001")},
            "00206-kwh": Decimal(cons),
        },
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MTR_UTLBIL_multiple_charges_one_tpr(mocker):
    MockSupply = mocker.patch("chellow.e.bill_parsers.haven_edi.Supply", autospec=True)
    mock_supply = mocker.Mock()
    MockSupply.get_by_mpan_core.return_value = mock_supply
    mock_era = mocker.Mock()
    mock_era.ssc.code = "0244"
    gbp = "10.31"
    cons = "113"
    mock_supply.find_era_at.return_value = mock_era
    elements = {}
    sess = mocker.Mock()
    headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
        ],
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "Day",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            },
        ],
        "bill_type_code": "N",
    }
    expected_headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
        ],
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {
            "00206-gbp": Decimal(gbp) * 2,
            "00206-rate": {Decimal("0.0001")},
            "00206-kwh": Decimal(cons) * 2,
        },
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MTR_UTLBIL_unmetered(mocker):
    MockSupply = mocker.patch("chellow.e.bill_parsers.haven_edi.Supply", autospec=True)
    mock_supply = mocker.Mock()
    MockSupply.get_by_mpan_core.return_value = mock_supply
    mock_era = mocker.Mock()
    mock_era.ssc.code = "0428"
    mock_mr_00258 = mocker.Mock()
    mock_mr_00258.tpr.code = "00258"
    mock_mr_00259 = mocker.Mock()
    mock_mr_00259.tpr.code = "00259"
    mock_era.ssc.measurement_requirements = [mock_mr_00258, mock_mr_00259]
    gbp = "10.31"
    cons = "113"
    mock_supply.find_era_at.return_value = mock_era
    elements = {}
    sess = mocker.Mock()
    headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Energy Charges",
            )
        ],
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [],
        "bill_type_code": "N",
    }
    expected_headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Energy Charges",
            )
        ],
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {
            "00258-gbp": Decimal(gbp) / 2,
            "00258-rate": {Decimal("0.0001")},
            "00258-kwh": Decimal(cons) / 2,
            "00259-gbp": Decimal(gbp) / 2,
            "00259-rate": {Decimal("0.0001")},
            "00259-kwh": Decimal(cons) / 2,
        },
        "reads": [],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MAN(mocker):
    elements = {
        "MADN": ["20", "0000000000", "6", "00", "001", "002"],
    }

    headers = {}
    _process_MAN(elements, headers)
    expected_headers = {"mpan_core": "20 0000 0000 006"}
    assert headers == expected_headers


def test_process_MHD(mocker):
    message_type = "UTLBIL"
    elements = {"TYPE": [message_type]}

    sess = mocker.Mock()
    headers = {"sess": sess}
    _process_MHD(elements, headers)
    expected_headers = {
        "message_type": message_type,
        "reads": [],
        "errors": [],
        "bill_elements": [],
        "breakdown": {"raw-lines": []},
        "sess": sess,
        "kwh": Decimal("0"),
    }
    assert headers == expected_headers
    assert type(headers["breakdown"]) is type(expected_headers)


def test_process_CCD3(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "CONS": [[]],
        "BPRI": ["10"],
    }

    headers = {
        "bill_elements": [],
        "kwh": Decimal("0"),
    }

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "bill_elements": [
            BillElement(
                gbp=Decimal("0.00"),
                rate=Decimal("0.0001"),
                cons=Decimal("0"),
                titles=None,
                desc="Night",
            )
        ],
    }
    assert headers == expected_headers


def test_process_CCD_1(mocker):
    msn = "hgkh"

    elements = {
        "CCDE": ["1", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "MTNR": [msn],
        "CONS": [[]],
        "BPRI": ["10"],
        "PRDT": ["200301"],
        "PVDT": ["200331"],
        "MLOC": [""],
        "PRRD": ["0", "00", "1", "00"],
        "ADJF": ["", "1"],
    }

    headers = {
        "reads": [],
    }

    _process_CCD1(elements, headers)

    expected_headers = {
        "reads": [
            {
                "msn": msn,
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "453043",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ]
    }
    assert headers == expected_headers


def test_process_CLO(mocker):
    account = "accnt"

    elements = {"CLOC": [account, ""]}

    headers = {}

    _process_CLO(elements, headers)

    expected_headers = {
        "account": "",
    }
    assert headers == expected_headers


def test_process_segment_error(mocker):
    mocker.patch(
        "chellow.e.bill_parsers.haven_edi._process_MTR", side_effect=BadRequest()
    )

    code = "MTR"
    elements = []
    line = ""
    account = "a1"
    bill_type_code = "N"
    breakdown = {"raw-lines": []}
    finish_date = utc_datetime(2020, 3, 31, 23, 30)
    gross = 1
    kwh = 1
    issue_date = utc_datetime(2020, 4, 30, 23, 30)
    net = 1
    reads = []
    reference = "20848747847"
    start_date = utc_datetime(2020, 3, 1)
    vat = 1
    headers = {
        "message_type": "UTLBIL",
        "errors": [],
        "type_code": "N",
        "reads": reads,
        "bill_type_code": "N",
        "breakdown": breakdown,
        "gross": gross,
        "vat": vat,
        "net": net,
        "kwh": kwh,
        "issue_date": issue_date,
        "finish_date": finish_date,
        "start_date": start_date,
        "account": account,
        "reference": reference,
    }
    line_number = 13

    bill = _process_segment(code, elements, line, headers, line_number)

    expected_bill = {
        "bill_type_code": bill_type_code,
        "account": account,
        "error": "Can't parse the line number 13 : The browser (or proxy) sent a "
        "request that this server could not understand. The key mpan_core is missing "
        "from the headers at line number 13.",
        "breakdown": breakdown,
        "finish_date": finish_date,
        "issue_date": issue_date,
        "gross": gross,
        "kwh": kwh,
        "net": net,
        "reads": reads,
        "reference": reference,
        "start_date": start_date,
        "vat": vat,
    }

    assert bill == expected_bill


def test_Parser(mocker, sess):
    file_lines = [
        "MHD=2+UTLBIL:3'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()



import csv
import sys
import threading
import traceback

from dateutil.relativedelta import relativedelta

from flask import g

from sqlalchemy.sql.expression import false, select, true

from chellow.dloads import open_file
from chellow.models import (
    Channel,
    Contract,
    Era,
    Session,
    Site,
    SiteEra,
    Snag,
    Supply,
    User,
)
from chellow.utils import csv_make_val, hh_before, req_bool, req_int, utc_datetime_now
from chellow.views import chellow_redirect


def content(contract_id, days_hidden, is_ignored, user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("channel_snags.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            titles = (
                "Hidden Days",
                "Chellow Id",
                "Imp MPAN Core",
                "Exp MPAN Core",
                "Site Code",
                "Site Name",
                "Snag Description",
                "Import Related?",
                "Channel Type",
                "Start Date",
                "Finish Date",
                "Is Ignored?",
                "Days Since Snag Finished",
                "Duration Of Snag (Days)",
            )
            writer.writerow(titles)

            contract = Contract.get_dc_by_id(sess, contract_id)

            now = utc_datetime_now()
            cutoff_date = now - relativedelta(days=days_hidden)
            q = (
                select(Snag, Channel, Era, Supply, SiteEra, Site)
                .join(Channel, Snag.channel_id == Channel.id)
                .join(Era, Channel.era_id == Era.id)
                .join(Supply, Era.supply_id == Supply.id)
                .join(SiteEra, Era.site_eras)
                .join(Site, SiteEra.site_id == Site.id)
                .where(
                    SiteEra.is_physical == true(),
                    Era.dc_contract == contract,
                    Snag.start_date < cutoff_date,
                )
                .order_by(
                    Site.code,
                    Supply.id,
                    Channel.imp_related,
                    Channel.channel_type,
                    Snag.description,
                    Snag.start_date,
                    Snag.id,
                )
            )
            if not is_ignored:
                q = q.where(Snag.is_ignored == false())

            for snag, channel, era, supply, site_era, site in sess.execute(q):
                snag_start = snag.start_date
                snag_finish = snag.finish_date
                imp_mc = "" if era.imp_mpan_core is None else era.imp_mpan_core
                exp_mc = "" if era.exp_mpan_core is None else era.exp_mpan_core

                if snag_finish is None:
                    duration = now - snag_start
                    age_of_snag = None
                else:
                    duration = snag_finish - snag_start
                    if hh_before(cutoff_date, snag_finish):
                        age_of_snag = None
                    else:
                        delta = now - snag_finish
                        age_of_snag = delta.days

                vals = {
                    "Hidden Days": days_hidden,
                    "Chellow Id": snag.id,
                    "Imp MPAN Core": imp_mc,
                    "Exp MPAN Core": exp_mc,
                    "Site Code": site.code,
                    "Site Name": site.name,
                    "Snag Description": snag.description,
                    "Import Related?": channel.imp_related,
                    "Channel Type": channel.channel_type,
                    "Start Date": snag_start,
                    "Finish Date": snag_finish,
                    "Is Ignored?": snag.is_ignored,
                    "Days Since Snag Finished": age_of_snag,
                    "Duration Of Snag (Days)": duration.days,
                }

                writer.writerow(csv_make_val(vals[t]) for t in titles)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    contract_id = req_int("dc_contract_id")
    days_hidden = req_int("days_hidden")
    is_ignored = req_bool("is_ignored")

    args = contract_id, days_hidden, is_ignored, g.user.id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import sys
import threading
import traceback
from collections import defaultdict

from dateutil.relativedelta import relativedelta

from flask import flash, g, make_response, render_template, request

import odio

from sqlalchemy import or_, select, true
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import contract_func, forecast_date
from chellow.e.scenario import make_calcs, make_site_deltas
from chellow.models import (
    Bill,
    Contract,
    Era,
    MeasurementRequirement,
    RSession,
    Scenario,
    Site,
    SiteEra,
    Source,
    Ssc,
    Supply,
    Tpr,
    User,
)
from chellow.utils import (
    HH,
    PropDict,
    c_months_u,
    ct_datetime,
    ct_datetime_now,
    hh_format,
    hh_min,
    hh_range,
    make_val,
    parse_mpan_core,
    req_bool,
    req_hh_date,
    req_int,
    req_str,
    to_ct,
    to_utc,
    utc_datetime_now,
)
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, "unmetered": 1, "nhh": 2, "amr": 3, "hh": 4}


def write_spreadsheet(fl, compressed, site_rows, supply_rows, era_rows, read_rows):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Supply Level", supply_rows)
        f.append_table("Era Level", era_rows)
        f.append_table("Normal Reads", read_rows)


def make_bill_row(titles, bill):
    return [bill.get(t) for t in titles]


def _add_bills(month_data, bills, chunk_start, chunk_finish):
    for bill in bills:
        bill_role_code = bill.batch.contract.market_role.code
        bill_start = bill.start_date
        bill_finish = bill.finish_date
        bill_duration = (bill_finish - bill_start).total_seconds() + (30 * 60)
        overlap_duration = (
            min(bill_finish, chunk_finish) - max(bill_start, chunk_start)
        ).total_seconds() + (30 * 60)
        proportion = overlap_duration / bill_duration
        bill_prop_kwh = proportion * float(bill.kwh)
        bill_prop_net_gbp = proportion * float(bill.net)
        bill_prop_vat_gbp = proportion * float(bill.vat)
        bill_prop_gross_gbp = proportion * float(bill.gross)
        if bill_role_code == "X":
            role_name = "supplier"
        elif bill_role_code == "C":
            role_name = "dc"
        elif bill_role_code == "M":
            role_name = "mop"
        else:
            raise BadRequest("Role code not recognized.")

        month_data["billed-import-kwh"] += bill_prop_kwh
        month_data["billed-import-net-gbp"] += bill_prop_net_gbp
        month_data[f"billed-import-{role_name}-kwh"] += bill_prop_kwh
        month_data[f"billed-import-{role_name}-net-gbp"] += bill_prop_net_gbp
        month_data[f"billed-import-{role_name}-vat-gbp"] += bill_prop_vat_gbp
        month_data[f"billed-import-{role_name}-gross-gbp"] += bill_prop_gross_gbp


def _process_site(
    sess,
    report_context,
    forecast_from,
    start_date,
    finish_date,
    site,
    scenario_props,
    supply_ids,
    now,
    summary_titles,
    title_dict,
    era_titles,
    era_rows,
    supply_titles,
    supply_rows,
    site_rows,
    data_source_bill,
):
    scenario_hh = scenario_props.get("hh_data", {})
    era_maps = scenario_props.get("era_maps", {})

    site_deltas = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_ids
    )

    calcs, site_gen_types = make_calcs(
        sess,
        site,
        start_date,
        finish_date,
        supply_ids,
        site_deltas,
        forecast_from,
        report_context,
        era_maps,
        data_source_bill,
    )

    site_category = None
    site_sources = set()
    normal_reads = set()
    site_data = defaultdict(int)
    supplies_data = {}
    for order, imp_mpan_core, exp_mpan_core, imp_ss, exp_ss in sorted(calcs, key=str):
        main_ss = exp_ss if imp_ss is None else imp_ss

        vals = {
            "creation-date": now,
            "start-date": main_ss.start_date,
            "finish-date": main_ss.finish_date,
            "site-code": site.code,
            "site-name": site.name,
        }
        for sname in (
            "import-grid",
            "export-grid",
            "import-gen",
            "export-gen",
            "import-3rd-party",
            "export-3rd-party",
            "msp",
            "used",
            "used-3rd-party",
        ):
            for xname in ("kwh", "net-gbp"):
                vals[f"{sname}-{xname}"] = 0
        for suf in ("kwh", "net-gbp", "vat-gbp", "gross-gbp"):
            vals[f"billed-import-{suf}"] = 0

        if imp_mpan_core == "displaced":
            try:
                supply_data = supplies_data["displaced"]
            except KeyError:
                supply_data = supplies_data[-1] = defaultdict(
                    int,
                    {
                        "creation-date": now,
                        "start-date": main_ss.start_date,
                        "finish-date": main_ss.finish_date,
                        "imp-mpan-core": "displaced",
                        "site-code": site.code,
                        "site-name": site.name,
                        "associated-site-codes": set(),
                        "source": "displaced",
                        "supply-name": "displaced",
                    },
                )

            vals["billed-import-supplier-net-gbp"] = None
            vals["billed-import-supplier-vat-gbp"] = None
            vals["billed-import-supplier-gross-gbp"] = None
            vals["billed-import-dc-net-gbp"] = None
            vals["billed-import-dc-vat-gbp"] = None
            vals["billed-import-dc-gross-gbp"] = None
            vals["billed-import-mop-net-gbp"] = None
            vals["billed-import-mop-vat-gbp"] = None
            vals["billed-import-mop-gross-gbp"] = None

            vals["used-kwh"] = vals["displaced-kwh"] = sum(
                hh["msp-kwh"] for hh in imp_ss.hh_data
            )

            disp_supplier_bill = imp_ss.supplier_bill

            gbp = disp_supplier_bill.get("net-gbp", 0)

            vals["used-net-gbp"] = vals["displaced-net-gbp"] = gbp

            vals["imp-supplier-contract"] = imp_ss.supplier_contract.name
            vals["metering-type"] = imp_ss.era.meter_category
            vals["source"] = "displaced"

            for t in title_dict["imp-supplier"]:
                try:
                    vals[t] = disp_supplier_bill[t]
                except KeyError:
                    pass

        else:
            source_code = main_ss.source_code
            supply = main_ss.supply

            try:
                supply_data = supplies_data[supply.id]
            except KeyError:
                supply_data = supplies_data[supply.id] = defaultdict(
                    int,
                    {
                        "creation-date": now,
                        "start-date": main_ss.start_date,
                        "finish-date": main_ss.finish_date,
                        "site-code": site.code,
                        "site-name": site.name,
                        "associated-site-codes": set(),
                    },
                )

            site_sources.add(source_code)

            for suf in ("net-gbp", "vat-gbp", "gross-gbp"):
                vals[f"billed-import-supplier-{suf}"] = 0
                vals[f"billed-import-dc-{suf}"] = 0
                vals[f"billed-import-mop-{suf}"] = 0

            if imp_ss is None:
                imp_bad_hhs = imp_bad_kwh = imp_supplier_contract_name = None
            else:
                imp_supplier_contract = imp_ss.supplier_contract
                imp_supplier_bill = imp_ss.supplier_bill
                if imp_supplier_contract is None:
                    imp_supplier_contract_name = None
                else:
                    imp_supplier_contract_name = imp_supplier_contract.name
                imp_bad_hhs = imp_bad_kwh = imp_md_kw = imp_md_kva = 0
                for hh in imp_ss.hh_data:
                    imp_md_kw = max(imp_md_kw, hh["msp-kw"])
                    imp_md_kva = max(imp_md_kva, hh["msp-kva"])
                    if hh["status"] != "A":
                        imp_bad_hhs += 1
                        imp_bad_kwh += hh["msp-kwh"]

                supply_data["imp-non-actual-hhs"] += imp_bad_hhs
                supply_data["imp-non-actual-kwh"] += imp_bad_kwh
                supply_data["imp-md-kw"] += imp_md_kw
                supply_data["imp-md-kva"] += imp_md_kva

                for t in title_dict["imp-supplier"]:
                    try:
                        vals[f"imp-supplier-{t}"] = imp_supplier_bill[t]
                    except KeyError:
                        pass

                for n in imp_ss.normal_reads:
                    normal_reads.add((imp_mpan_core, n))

                kwh = sum(hh["msp-kwh"] for hh in imp_ss.hh_data)

                gbp = imp_supplier_bill.get("net-gbp", 0)

                if source_code in ("grid", "gen-grid"):
                    vals["import-grid-net-gbp"] += gbp
                    vals["import-grid-kwh"] += kwh
                    vals["used-net-gbp"] += gbp
                    vals["used-kwh"] += kwh
                    if source_code == "gen-grid":
                        vals["export-gen-kwh"] += kwh
                elif source_code == "3rd-party":
                    vals["import-3rd-party-net-gbp"] += gbp
                    vals["import-3rd-party-kwh"] += kwh
                    vals["used-3rd-party-net-gbp"] += gbp
                    vals["used-3rd-party-kwh"] += kwh
                    vals["used-net-gbp"] += gbp
                    vals["used-kwh"] += kwh
                elif source_code == "3rd-party-reverse":
                    vals["export-3rd-party-net-gbp"] += gbp
                    vals["export-3rd-party-kwh"] += kwh
                    vals["used-3rd-party-net-gbp"] -= gbp
                    vals["used-3rd-party-kwh"] -= kwh
                    vals["used-net-gbp"] -= gbp
                    vals["used-kwh"] -= kwh
                elif source_code == "gen":
                    vals["import-gen-kwh"] += kwh

            if exp_ss is None:
                exp_bad_hhs = exp_bad_kwh = exp_supplier_contract_name = None
            else:
                exp_supplier_contract = exp_ss.supplier_contract
                exp_supplier_bill = exp_ss.supplier_bill
                if exp_supplier_contract is None:
                    exp_supplier_contract_name = ""
                else:
                    exp_supplier_contract_name = exp_supplier_contract.name
                exp_bad_hhs = exp_bad_kwh = exp_md_kw = exp_md_kva = 0
                for hh in exp_ss.hh_data:
                    exp_md_kw = max(exp_md_kw, hh["msp-kw"])
                    exp_md_kva = max(exp_md_kva, hh["msp-kva"])
                    if hh["status"] != "A":
                        exp_bad_hhs += 1
                        exp_bad_kwh += hh["msp-kwh"]

                supply_data["exp-non-actual-hhs"] += exp_bad_hhs
                supply_data["exp-non-actual-kwh"] += exp_bad_kwh
                supply_data["exp-md-kw"] += exp_md_kw
                supply_data["exp-md-kva"] += exp_md_kva

                for t in title_dict["exp-supplier"]:
                    try:
                        vals[f"exp-supplier-{t}"] = exp_supplier_bill[t]
                    except KeyError:
                        pass

                kwh = sum(hh["msp-kwh"] for hh in exp_ss.hh_data)
                gbp = exp_supplier_bill.get("net-gbp", 0)

                if source_code in ("net", "gen-net"):
                    vals["export-grid-net-gbp"] += gbp
                    vals["export-grid-net-kwh"] += kwh
                    if source_code == "gen-net":
                        vals["import-gen-kwh"] += kwh

                elif source_code == "3rd-party":
                    vals["export-3rd-party-net-gbp"] += gbp
                    vals["export-3rd-party-kwh"] += kwh
                    vals["used-3rd-party-net-gbp"] -= gbp
                    vals["used-3rd-party-kwh"] -= kwh
                    vals["used-net-gbp"] -= gbp
                    vals["used-kwh"] -= kwh
                elif source_code == "3rd-party-reverse":
                    vals["import-3rd-party-net-gbp"] += gbp
                    vals["import-3rd-party-kwh"] += kwh
                    vals["used-3rd-party-net-gbp"] += gbp
                    vals["used-3rd-party-kwh"] += kwh
                    vals["used-net-gbp"] += gbp
                    vals["used-kwh"] += kwh
                elif source_code == "gen":
                    vals["export-gen-kwh"] += kwh

            sss = exp_ss if imp_ss is None else imp_ss
            dc_contract = sss.dc_contract
            dc_vb_func = sss.contract_func(dc_contract, "virtual_bill")
            try:
                dc_vb_func(sss)
            except AttributeError as e:
                raise BadRequest(
                    f"Problem with virtual bill of DC contract {dc_contract.id} {e} "
                    f"{traceback.format_exc()}"
                )
            dc_bill = sss.dc_bill
            gbp = dc_bill["net-gbp"]

            mop_contract = sss.mop_contract
            mop_bill_function = sss.contract_func(mop_contract, "virtual_bill")
            try:
                mop_bill_function(sss)
            except (AttributeError, NameError) as e:
                raise BadRequest(
                    f"Problem with virtual bill of MOP contract {mop_contract.id} {e} "
                    f"{traceback.format_exc()}"
                )
            mop_bill = sss.mop_bill
            gbp += mop_bill["net-gbp"]

            if source_code in ("3rd-party", "3rd-party-reverse"):
                vals["import-3rd-party-net-gbp"] += gbp
                vals["used-3rd-party-net-gbp"] += gbp
            else:
                vals["import-grid-net-gbp"] += gbp
            vals["used-net-gbp"] += gbp

            generator_type = sss.generator_type_code
            if source_code in ("gen", "gen-grid"):
                site_gen_types.add(generator_type)

            era_category = sss.measurement_type
            if CATEGORY_ORDER[site_category] < CATEGORY_ORDER[era_category]:
                site_category = era_category

            era_associates = {
                s.site.code for s in sss.era.site_eras if not s.is_physical
            }

            for bill in sess.query(Bill).filter(
                Bill.supply == supply,
                Bill.start_date <= sss.finish_date,
                Bill.finish_date >= sss.start_date,
            ):
                bill_role_code = bill.batch.contract.market_role.code
                bill_start = bill.start_date
                bill_finish = bill.finish_date
                bill_duration = (bill_finish - bill_start).total_seconds() + (30 * 60)
                overlap_duration = (
                    min(bill_finish, sss.finish_date) - max(bill_start, sss.start_date)
                ).total_seconds() + (30 * 60)
                proportion = overlap_duration / bill_duration
                vals["billed-import-kwh"] += proportion * float(bill.kwh)
                bill_prop_net_gbp = proportion * float(bill.net)
                bill_prop_vat_gbp = proportion * float(bill.vat)
                bill_prop_gross_gbp = proportion * float(bill.gross)
                vals["billed-import-net-gbp"] += bill_prop_net_gbp
                vals["billed-import-vat-gbp"] += bill_prop_vat_gbp
                vals["billed-import-gross-gbp"] += bill_prop_gross_gbp
                if bill_role_code == "X":
                    role_name = "supplier"
                elif bill_role_code == "C":
                    role_name = "dc"
                elif bill_role_code == "M":
                    role_name = "mop"
                else:
                    raise BadRequest("Role code not recognized.")

                vals[f"billed-import-{role_name}-net-gbp"] += bill_prop_net_gbp
                vals[f"billed-import-{role_name}-vat-gbp"] += bill_prop_vat_gbp
                vals[f"billed-import-{role_name}-gross-gbp"] += bill_prop_gross_gbp

            vals["imp-mpan-core"] = imp_mpan_core
            vals["exp-mpan-core"] = exp_mpan_core
            vals["associated-site-codes"] = era_associates
            vals["era-start-date"] = sss.era.start_date
            vals["era-finish-date"] = sss.era.finish_date
            vals["imp-supplier-contract"] = imp_supplier_contract_name
            vals["imp-non-actual-hhs"] = imp_bad_hhs
            vals["imp-non-actual-kwh"] = imp_bad_kwh
            vals["exp-supplier-contract"] = exp_supplier_contract_name
            vals["exp-non-actual-hhs"] = exp_bad_hhs
            vals["exp-non-actual-kwh"] = exp_bad_kwh
            vals["metering-type"] = era_category
            vals["source"] = source_code
            vals["generator-type"] = generator_type
            vals["supply-name"] = sss.supply_name
            vals["msn"] = sss.msn
            vals["pc"] = sss.pc_code

            for t in title_dict["mop"]:
                try:
                    vals[f"mop-{t}"] = mop_bill[t]
                except KeyError:
                    pass

            for t in title_dict["dc"]:
                try:
                    vals[f"dc-{t}"] = dc_bill[t]
                except KeyError:
                    pass

            supply_data["imp-mpan-core"] = imp_mpan_core
            supply_data["exp-mpan-core"] = exp_mpan_core
            supply_data["associated-site-codes"].update(era_associates)
            supply_data["start-date"] = min(sss.start_date, supply_data["start-date"])
            supply_data["finish-date"] = max(
                sss.finish_date, supply_data["finish-date"]
            )

            supply_data["source"] = source_code
            supply_data["generator-type"] = generator_type
            supply_data["supply-name"] = sss.supply_name

        era_rows.append([make_val(vals.get(t)) for t in era_titles])

        for t in summary_titles:
            v = vals.get(t)
            if v is not None:
                site_data[t] += v

    q = select(Supply).join(Era).join(SiteEra).where(SiteEra.site == site).distinct()
    if supply_ids is not None:
        q = q.where(Supply.id.in_(supply_ids))
    for supply in sess.execute(q).scalars():
        last_era = (
            sess.execute(
                select(Era).where(Era.supply == supply).order_by(Era.start_date.desc())
            )
            .scalars()
            .first()
        )
        if last_era.finish_date is not None and last_era.start_date <= finish_date:
            site_era = sess.execute(
                select(SiteEra).where(
                    SiteEra.era == last_era,
                    SiteEra.is_physical == true(),
                    SiteEra.site == site,
                )
            ).scalar_one_or_none()
            if site_era is not None:
                chunk_start = max(start_date, last_era.finish_date + HH)
                chunk_finish = finish_date
                bills = (
                    sess.execute(
                        select(Bill).where(
                            Bill.supply == supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                    )
                    .scalars()
                    .all()
                )
                if len(bills) > 0:
                    imp_supplier_contract = last_era.imp_supplier_contract
                    imp_supplier_contract_name = (
                        None
                        if imp_supplier_contract is None
                        else imp_supplier_contract.name
                    )
                    exp_supplier_contract = last_era.exp_supplier_contract
                    exp_supplier_contract_name = (
                        None
                        if exp_supplier_contract is None
                        else exp_supplier_contract.name
                    )
                    era_associates = {
                        s.site.code for s in last_era.site_eras if not s.is_physical
                    }
                    vals = {
                        "creation-date": now,
                        "start-date": chunk_start,
                        "finish-date": chunk_finish,
                        "imp-mpan-core": last_era.imp_mpan_core,
                        "exp-mpan-core": last_era.exp_mpan_core,
                        "site-code": site.code,
                        "site-name": site.name,
                        "associated-site-codes": era_associates,
                        "era-start-date": None,
                        "era-finish-date": None,
                        "imp-supplier-contract": imp_supplier_contract_name,
                        "imp-non-actual-hhs": None,
                        "imp-non-actual-kwh": None,
                        "exp-supplier-contract": exp_supplier_contract_name,
                        "exp-non-actual-hhs": None,
                        "exp-non-actual-kwh": None,
                        "metering-type": last_era.meter_category,
                        "source": last_era.supply.source.code,
                        "generator-type": None,
                        "supply-name": last_era.supply.name,
                        "msn": last_era.msn,
                        "pc": last_era.pc.code,
                    }
                    for name in (
                        "import-grid",
                        "export-grid",
                        "import-gen",
                        "export-gen",
                        "import-3rd-party",
                        "export-3rd-party",
                        "displaced",
                        "used",
                        "used-3rd-party",
                    ):
                        for sname in ("kwh", "net-gbp"):
                            vals[f"{name}-{sname}"] = 0
                    for suf in ("kwh", "net-gbp", "vat-gbp", "gross-gbp"):
                        vals[f"billed-import-{suf}"] = 0
                        vals[f"billed-import-supplier-{suf}"] = 0
                        vals[f"billed-import-dc-{suf}"] = 0
                        vals[f"billed-import-mop-{suf}"] = 0

                    _add_bills(vals, bills, chunk_start, chunk_finish)

                    era_rows.append([make_val(vals.get(t)) for t in era_titles])
                    for t in summary_titles:
                        v = vals.get(t)
                        if v is not None:
                            site_data[t] += v
        first_era = (
            sess.execute(
                select(Era).where(Era.supply == supply).order_by(Era.start_date)
            )
            .scalars()
            .first()
        )
        if first_era.start_date > start_date:
            site_era = sess.execute(
                select(SiteEra).where(
                    SiteEra.era == first_era,
                    SiteEra.is_physical == true(),
                    SiteEra.site == site,
                )
            ).scalar_one_or_none()
            if site_era is not None:
                chunk_start = start_date
                chunk_finish = hh_min(finish_date, first_era.start_date - HH)
                bills = (
                    sess.execute(
                        select(Bill).where(
                            Bill.supply == supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                    )
                    .scalars()
                    .all()
                )
                imp_supplier_contract = first_era.imp_supplier_contract
                exp_supplier_contract = first_era.exp_supplier_contract
                if len(bills) > 0:
                    era_vals = {
                        "creation-date": now,
                        "start-date": start_date,
                        "finish-date": finish_date,
                        "imp-mpan-core": last_era.imp_mpan_core,
                        "exp-mpan-core": last_era.exp_mpan_core,
                        "site-code": site.code,
                        "site-name": site.name,
                        "associated-site-codes": None,
                        "era-start-date": None,
                        "era-finish-date": None,
                        "imp-supplier-contract": (
                            None
                            if imp_supplier_contract is None
                            else imp_supplier_contract.name
                        ),
                        "imp-non-actual-hhs": None,
                        "imp-non-actual-kwh": None,
                        "exp-supplier-contract": (
                            None
                            if exp_supplier_contract is None
                            else exp_supplier_contract.name
                        ),
                        "exp-non-actual-hhs": None,
                        "exp-non-actual-kwh": None,
                        "metering-type": last_era.meter_category,
                        "source": last_era.supply.source.code,
                        "generator-type": None,
                        "supply-name": last_era.supply.name,
                        "msn": last_era.msn,
                        "pc": last_era.pc.code,
                    }
                    for name in (
                        "import-grid",
                        "export-grid",
                        "import-gen",
                        "export-gen",
                        "import-3rd-party",
                        "export-3rd-party",
                        "displaced",
                        "used",
                        "used-3rd-party",
                    ):
                        for sname in ("kwh", "net-gbp"):
                            era_vals[f"{name}-{sname}"] = 0
                    for suf in ("kwh", "net-gbp", "vat-gbp", "gross-gbp"):
                        era_vals[f"billed-import-{suf}"] = 0
                        era_vals[f"billed-import-supplier-{suf}"] = 0
                        era_vals[f"billed-import-dc-{suf}"] = 0
                        era_vals[f"billed-import-mop-{suf}"] = 0

                    _add_bills(era_vals, bills, chunk_start, chunk_finish)

                    era_rows.append([make_val(era_vals.get(t)) for t in era_titles])
                    for t in summary_titles:
                        v = era_vals.get(t)
                        if v is not None:
                            site_data[t] += v

    for _, v in sorted(supplies_data.items()):
        row = [make_val(v.get(t)) for t in supply_titles]
        supply_rows.append(row)

    site_md_used_kw = 0
    for hh in site.hh_data(sess, start_date, finish_date, exclude_virtual=True):
        used_kw = hh["used"] * 2
        site_md_used_kw = max(used_kw, site_md_used_kw)

    site_row = [
        now,
        site.code,
        site.name,
        [s.code for s in site.find_linked_sites(sess, start_date, finish_date)],
        start_date,
        finish_date,
        site_category,
        site_sources,
        site_gen_types,
        site_md_used_kw,
    ] + [site_data[k] for k in summary_titles]

    site_rows.append([make_val(v) for v in site_row])
    return normal_reads


class Object:
    pass


def content(
    scenario_props,
    base_name,
    user_id,
    compression,
    now,
    is_bill_check,
):
    report_context = {}

    try:
        comp = report_context["computer"]
    except KeyError:
        comp = report_context["computer"] = {}

    try:
        rate_cache = comp["rates"]
    except KeyError:
        rate_cache = comp["rates"] = {}

    try:
        ind_cont = report_context["contract_names"]
    except KeyError:
        ind_cont = report_context["contract_names"] = {}

    rsess = rf = None
    site_rows = []
    supply_rows = []
    era_rows = []
    normal_read_rows = []

    try:
        with RSession() as rsess:
            start_date = to_utc(
                ct_datetime(
                    scenario_props["scenario_start_year"],
                    scenario_props["scenario_start_month"],
                    scenario_props["scenario_start_day"],
                    scenario_props["scenario_start_hour"],
                    scenario_props["scenario_start_minute"],
                )
            )
            finish_date = to_utc(
                ct_datetime(
                    scenario_props["scenario_finish_year"],
                    scenario_props["scenario_finish_month"],
                    scenario_props["scenario_finish_day"],
                    scenario_props["scenario_finish_hour"],
                    scenario_props["scenario_finish_minute"],
                )
            )

            if "forecast_from" in scenario_props:
                forecast_from = scenario_props["forecast_from"]
            else:
                forecast_from = None

            if forecast_from is None:
                forecast_from = forecast_date()
            else:
                forecast_from = to_utc(forecast_from)

            site_codes = scenario_props.get("site_codes")

            sites = select(Site).distinct().order_by(Site.code)
            if site_codes is not None and len(site_codes) > 0:
                base_name.append("sitecodes")
                sites = sites.where(Site.code.in_(site_codes))

            mpan_cores = scenario_props.get("mpan_cores")
            if mpan_cores is None:
                supply_ids = None
            else:
                supply_ids = [Supply.get_by_mpan_core(rsess, m).id for m in mpan_cores]
                sites = (
                    sites.join(SiteEra).join(Era).where(Era.supply_id.in_(supply_ids))
                )

            user = User.get_by_id(rsess, user_id)
            rf = open_file("_".join(base_name) + ".ods", user, mode="wb")

            for rate_script in scenario_props.get("rates", []):
                contract_id = rate_script["contract_id"]
                try:
                    cont_cache = rate_cache[contract_id]
                except KeyError:
                    cont_cache = rate_cache[contract_id] = {}

                try:
                    rate_script_start = rate_script["start_date"]
                except KeyError:
                    raise BadRequest(
                        f"Problem in the scenario properties. Can't find the "
                        f"'start_date' key of the contract {contract_id} in the "
                        f"'rates' map."
                    )

                try:
                    rate_script_start = rate_script["start_date"]
                except KeyError:
                    raise BadRequest(
                        f"Problem in the scenario properties. Can't find the "
                        f"'start_date' key of the contract {contract_id} in the "
                        f"'rates' map."
                    )

                props = PropDict("scenario properties", rate_script["script"])
                for dt in hh_range(
                    report_context, rate_script_start, rate_script["finish_date"]
                ):
                    cont_cache[dt] = props

            for rate_script in scenario_props.get("industry_rates", []):
                contract_name = rate_script["contract_name"]
                try:
                    cont_cache = ind_cont[contract_name]
                except KeyError:
                    cont_cache = ind_cont[contract_name] = {}

                rfinish = rate_script["finish_date"]
                if rfinish is None:
                    raise BadRequest(
                        f"For the industry rate {contract_name} the finish_date can't "
                        f"be null."
                    )
                for dt in hh_range(report_context, rate_script["start_date"], rfinish):
                    cont_cache[dt] = PropDict(
                        "scenario properties", rate_script["script"]
                    )

            by_hh = scenario_props.get("by_hh", False)
            is_bill_check = scenario_props.get("is_bill_check", False)

            era_header_titles = [
                "creation-date",
                "start-date",
                "finish-date",
                "imp-mpan-core",
                "exp-mpan-core",
                "site-code",
                "site-name",
                "associated-site-codes",
                "era-start-date",
                "era-finish-date",
                "imp-supplier-contract",
                "imp-non-actual-hhs",
                "imp-non-actual-kwh",
                "exp-supplier-contract",
                "exp-non-actual-hhs",
                "exp-non-actual-kwh",
                "metering-type",
                "source",
                "generator-type",
                "supply-name",
                "msn",
                "pc",
            ]
            supply_header_titles = [
                "creation-date",
                "start-date",
                "finish-date",
                "imp-mpan-core",
                "exp-mpan-core",
                "site-code",
                "site-name",
                "associated-site-codes",
                "source",
                "generator-type",
                "supply-name",
                "imp-md-kw",
                "imp-md-kva",
                "exp-md-kw",
                "exp-md-kva",
            ]
            site_header_titles = [
                "creation-date",
                "site-code",
                "site-name",
                "associated-site-codes",
                "start-date",
                "finish-date",
                "metering-type",
                "sources",
                "generator-types",
                "md-used-kw",
            ]
            summary_titles = [
                "import-grid-kwh",
                "export-grid-kwh",
                "import-gen-kwh",
                "export-gen-kwh",
                "import-3rd-party-kwh",
                "export-3rd-party-kwh",
                "displaced-kwh",
                "used-kwh",
                "used-3rd-party-kwh",
                "import-grid-net-gbp",
                "export-grid-net-gbp",
                "import-gen-net-gbp",
                "export-gen-net-gbp",
                "import-3rd-party-net-gbp",
                "export-3rd-party-net-gbp",
                "displaced-net-gbp",
                "used-net-gbp",
                "used-3rd-party-net-gbp",
                "billed-import-kwh",
                "billed-import-net-gbp",
                "billed-import-vat-gbp",
                "billed-import-gross-gbp",
                "billed-import-supplier-net-gbp",
                "billed-import-supplier-vat-gbp",
                "billed-import-supplier-gross-gbp",
                "billed-import-dc-net-gbp",
                "billed-import-dc-vat-gbp",
                "billed-import-dc-gross-gbp",
                "billed-import-mop-net-gbp",
                "billed-import-mop-vat-gbp",
                "billed-import-mop-gross-gbp",
            ]

            title_dict = {}
            for cont_type, con_attr in (
                ("mop", Era.mop_contract),
                ("dc", Era.dc_contract),
                ("imp-supplier", Era.imp_supplier_contract),
                ("exp-supplier", Era.exp_supplier_contract),
            ):
                titles = []
                title_dict[cont_type] = titles
                conts = (
                    select(Contract)
                    .join(con_attr)
                    .join(Era.supply)
                    .join(Source)
                    .where(
                        Era.start_date <= finish_date,
                        or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    )
                    .distinct()
                    .order_by(Contract.id)
                )
                if supply_ids is not None:
                    conts = conts.where(Era.supply_id.in_(supply_ids))
                for cont in rsess.scalars(conts):
                    title_func = contract_func(
                        report_context, cont, "virtual_bill_titles"
                    )
                    if title_func is None:
                        raise Exception(
                            f"For the contract {cont.name} there doesn't seem to be a "
                            f"'virtual_bill_titles' function."
                        )
                    for title in title_func():
                        if title not in titles:
                            titles.append(title)

            tpr_query = (
                rsess.query(Tpr)
                .join(MeasurementRequirement)
                .join(Ssc)
                .join(Era)
                .filter(
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
                .order_by(Tpr.code)
                .distinct()
            )
            for tpr in tpr_query.filter(Era.imp_supplier_contract != null()):
                for suffix in ("-kwh", "-rate", "-gbp"):
                    title_dict["imp-supplier"].append(tpr.code + suffix)
            for tpr in tpr_query.filter(Era.exp_supplier_contract != null()):
                for suffix in ("-kwh", "-rate", "-gbp"):
                    title_dict["exp-supplier"].append(tpr.code + suffix)

            era_titles = (
                era_header_titles
                + summary_titles
                + [None]
                + ["mop-" + t for t in title_dict["mop"]]
                + [None]
                + ["dc-" + t for t in title_dict["dc"]]
                + [None]
                + ["imp-supplier-" + t for t in title_dict["imp-supplier"]]
                + [None]
                + ["exp-supplier-" + t for t in title_dict["exp-supplier"]]
            )
            site_rows.append(site_header_titles + summary_titles)
            supply_titles = supply_header_titles + summary_titles
            supply_rows.append(supply_titles)
            era_rows.append(era_titles)

            normal_reads = set()

            if is_bill_check:
                data_source_bill = Object()
                data_source_bill.start_date = start_date
                data_source_bill.finish_date = finish_date
            else:
                data_source_bill = None

            for site in rsess.scalars(sites):
                if by_hh:
                    sf = [
                        (d, d)
                        for d in hh_range(report_context, start_date, finish_date)
                    ]
                else:
                    sf = [(start_date, finish_date)]

                for start, finish in sf:
                    try:
                        normal_reads = normal_reads | _process_site(
                            rsess,
                            report_context,
                            forecast_from,
                            start,
                            finish,
                            site,
                            scenario_props,
                            supply_ids,
                            now,
                            summary_titles,
                            title_dict,
                            era_titles,
                            era_rows,
                            supply_titles,
                            supply_rows,
                            site_rows,
                            data_source_bill,
                        )
                    except BadRequest as e:
                        raise BadRequest(f"Site Code {site.code}: {e.description}")

                normal_read_rows = [["mpan_core", "date", "msn", "type", "registers"]]
                for mpan_core, r in sorted(list(normal_reads)):
                    row = [mpan_core, r.date, r.msn, r.type] + list(r.reads)
                    normal_read_rows.append(row)

                rsess.rollback()  # Evict from cache
            write_spreadsheet(
                rf, compression, site_rows, supply_rows, era_rows, normal_read_rows
            )
    except BadRequest as e:
        msg = e.description + traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        write_spreadsheet(
            rf, compression, site_rows, supply_rows, era_rows, normal_read_rows
        )
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        if rf is None:
            msg = traceback.format_exc()
            ef = open_file("error.txt", None, mode="w")
            ef.write(msg + "\n")
            ef.close()
        else:
            write_spreadsheet(
                rf, compression, site_rows, supply_rows, era_rows, normal_read_rows
            )
    finally:
        if rf is not None:
            rf.close()


def do_post(sess):
    base_name = ["duration"]
    compression = req_bool("compression")

    try:
        if "scenario_id" in request.values:
            scenario_id = req_int("scenario_id")
            scenario = Scenario.get_by_id(sess, scenario_id)
            props = scenario.props
            base_name.append(scenario.name)

        else:
            start_date = req_hh_date("start")
            finish_date = req_hh_date("finish")

            start_date_ct = to_ct(start_date)
            finish_date_ct = to_ct(finish_date)
            props = {
                "scenario_start_year": start_date_ct.year,
                "scenario_start_month": start_date_ct.month,
                "scenario_start_day": start_date_ct.day,
                "scenario_start_hour": start_date_ct.hour,
                "scenario_start_minute": start_date_ct.minute,
                "scenario_finish_year": finish_date_ct.year,
                "scenario_finish_month": finish_date_ct.month,
                "scenario_finish_day": finish_date_ct.day,
                "scenario_finish_hour": finish_date_ct.hour,
                "scenario_finish_minute": finish_date_ct.minute,
            }
            base_name.append(hh_format(start_date).replace(" ", "_").replace(":", "_"))

        if "mpan_cores" in request.values:
            mpan_cores_str = req_str("mpan_cores")
            mpan_cores_lines = mpan_cores_str.splitlines()
            if len(mpan_cores_lines) > 0:
                props["mpan_cores"] = [parse_mpan_core(m) for m in mpan_cores_lines]

        if "supply_id" in request.values:
            supply_id = req_int("supply_id")
            supply = Supply.get_by_id(sess, supply_id)
            era = supply.eras[-1]
            imp_mpan_core, exp_mpan_core = era.imp_mpan_core, era.exp_mpan_core
            props["mpan_cores"] = [
                exp_mpan_core if imp_mpan_core is None else imp_mpan_core
            ]

        if "site_codes" in request.values:
            site_codes_str = req_str("site_codes")
            site_codes_lines = site_codes_str.splitlines()
            if len(site_codes_lines) > 0:
                props["site_codes"] = [c.strip() for c in site_codes_lines]

        if "site_id" in request.values:
            site_id = req_int("site_id")
            site = Site.get_by_id(sess, site_id)
            props["site_codes"] = [site.code]

        now = utc_datetime_now()
        args = props, base_name, g.user.id, compression, now, False
        sess.rollback()
        thread = threading.Thread(target=content, args=args)
        thread.start()
        return chellow_redirect("/downloads", 303)
    except BadRequest as e:
        flash(e.description)
        if "scenario_id" in request.values:
            return make_response(
                render_template("e/scenario.html", scenario=scenario), 400
            )
        else:
            ct_last = ct_datetime_now() - relativedelta(months=1)
            months = list(
                c_months_u(start_year=ct_last.year, start_month=ct_last.month)
            )
            month_start, month_finish = months[0]
            return make_response(
                render_template(
                    "e/duration_report.html",
                    month_start=month_start,
                    month_finish=month_finish,
                ),
                400,
            )



from datetime import datetime as Datetime
from decimal import Decimal

from dateutil.relativedelta import relativedelta

from sqlalchemy import null, or_, select

from werkzeug.exceptions import BadRequest

from chellow.e.lcc import api_records
from chellow.models import Contract, RateScript
from chellow.utils import c_months_u, ct_datetime, hh_format, to_ct, to_utc


def _find_quarter_rs(sess, contract_name, date):
    contract = Contract.get_non_core_by_name(sess, contract_name)
    rs = sess.scalars(
        select(RateScript).where(
            RateScript.contract == contract,
            RateScript.start_date <= date,
            or_(
                RateScript.finish_date == null(),
                RateScript.finish_date >= date,
            ),
        )
    ).one_or_none()
    if rs is None:
        return False
    else:
        st_ct = to_ct(rs.start_date)
        if (
            rs.finish_date is None
            and date
            > list(
                c_months_u(start_year=st_ct.year, start_month=st_ct.month, months=3)
            )[-1][-1]
        ):
            return False
        else:
            return True


def hh(data_source, use_bill_check=False):
    try:
        cfd_cache = data_source.caches["cfd"]
    except KeyError:
        cfd_cache = data_source.caches["cfd"] = {}

    for h in data_source.hh_data:
        try:
            h["cfd-rate"] = cfd_cache[h["start-date"]]
        except KeyError:
            h_start = h["start-date"]
            if use_bill_check:
                base_rate = (
                    float(
                        data_source.non_core_rate("cfd_forecast_ilr_tra", h_start)[
                            "record"
                        ]["Interim_Levy_Rate_GBP_Per_MWh"]
                    )
                    / 1000
                )
            else:
                if _find_quarter_rs(
                    data_source.sess, "cfd_reconciled_daily_levy_rates", h_start
                ):
                    base_rate_dec = data_source.non_core_rate(
                        "cfd_reconciled_daily_levy_rates", h_start
                    )["rate_gbp_per_kwh"]

                elif _find_quarter_rs(
                    data_source.sess, "cfd_in_period_tracking", h_start
                ):
                    base_rate_dec = data_source.non_core_rate(
                        "cfd_in_period_tracking", h_start
                    )["rate_gbp_per_kwh"]

                elif _find_quarter_rs(
                    data_source.sess, "cfd_forecast_ilr_tra", h_start
                ):
                    base_rate_dec = Decimal(
                        data_source.non_core_rate("cfd_forecast_ilr_tra", h_start)[
                            "record"
                        ]["Interim_Levy_Rate_GBP_Per_MWh"]
                    ) / Decimal(1000)

                else:
                    base_rate_dec = Decimal(
                        data_source.non_core_rate(
                            "cfd_advanced_forecast_ilr_tra", h_start
                        )["sensitivity"]["Base Case"]["Interim Levy Rate_GBP_Per_MWh"]
                    ) / Decimal(1000)

                base_rate = float(base_rate_dec)

            effective_ocl_rate = data_source.non_core_rate(
                "cfd_operational_costs_levy", h["start-date"]
            )["record"]["Effective_OCL_Rate_GBP_Per_MWh"]
            if effective_ocl_rate == "":
                levy_rate_str = data_source.non_core_rate(
                    "cfd_operational_costs_levy", h_start
                )["record"]["OCL_Rate_GBP_Per_MWh"]
            else:
                levy_rate_str = effective_ocl_rate
            levy_rate = float(levy_rate_str) / 1000

            h["cfd-rate"] = cfd_cache[h_start] = base_rate + levy_rate


def lcc_import(sess, log, set_progress, s):
    import_in_period_tracking(sess, log, set_progress, s)
    import_operational_costs_levy(sess, log, set_progress, s)
    import_reconciled_daily_levy_rates(sess, log, set_progress, s)
    import_forecast_ilr_tra(sess, log, set_progress, s)
    import_advanced_forecast_ilr_tra(sess, log, set_progress, s)


def _quarters(log, s):
    quarter = {}
    for record in api_records(log, s, "2fc2fad9-ad57-4901-982a-f92d4ef6c622"):
        settlement_date_str = record["Settlement_Date"]
        settlement_date_ct = to_ct(
            Datetime.strptime(settlement_date_str[:10], "%Y-%m-%d")
        )
        settlement_date = to_utc(settlement_date_ct)

        quarter[settlement_date] = record
        settlement_next_ct = settlement_date_ct + relativedelta(days=1)

        if settlement_next_ct.month in (1, 4, 7, 10) and settlement_next_ct.day == 1:
            yield quarter
            quarter = {}


def _parse_number(num_str):
    if num_str == "":
        return 0
    else:
        return float(num_str)


def _parse_date(date_str):
    return to_utc(to_ct(Datetime.strptime(date_str[:10], "%Y-%m-%d")))


def _parse_varying_date(date_str):
    if "-" in date_str:
        pattern = "%Y-%m-%d"
    elif "/" in date_str:
        pattern = "%d/%m/%Y"
    else:
        raise BadRequest(f"The date {date_str} is not recognized.")
    return to_utc(to_ct(Datetime.strptime(date_str[:10], pattern)))


def import_in_period_tracking(sess, log, set_progress, s):
    log("Starting to check for new LCC CfD In-Period Tracking")

    contract_name = "cfd_in_period_tracking"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for quarter in _quarters(log, s):
        quarter_start = sorted(quarter.keys())[0]
        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == quarter_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, quarter_start, {"records": {}})

        gbp = 0
        kwh = 0
        for record in quarter.values():
            gbp += _parse_number(record["Actual_CFD_Payments_GBP"])
            gbp += _parse_number(record["Expected_CFD_Payments_GBP"])
            kwh += _parse_number(record["Actual_Eligible_Demand_MWh"]) * 1000
            kwh += _parse_number(record["Expected_Eligible_Demand_MWh"]) * 1000

        rate = gbp / kwh

        rs_script = rs.make_script()
        records = rs_script["records"]
        for k, v in sorted(quarter.items()):
            records[hh_format(k)] = v
        rs_script["rate_gbp_per_kwh"] = rate
        rs.update(rs_script)
        sess.commit()
    log("Finished LCC CfD In-Period Tracking")


def import_operational_costs_levy(sess, log, set_progress, s):
    log("Starting to check for new LCC CfD Operational Costs Levy")

    contract_name = "cfd_operational_costs_levy"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for record in api_records(log, s, "44f41eac-61b3-4e8d-8c52-3eda7b8e8517", skip=1):
        period_start = _parse_date(record["Period_Start"])

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == period_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, period_start, {})

        rs_script = rs.make_script()
        rs_script["record"] = record
        rs.update(rs_script)
        sess.commit()


RUN_TYPES = ("II", "SF", "R1", "R2", "R3", "RF", "DF")


def _reconciled_quarters(log, s, search_from):
    quarters = {}

    for record in api_records(log, s, "e0e163cb-ba36-416d-83fe-976992d61516"):
        settlement_date = _parse_date(record["Settlement_Date"])
        if settlement_date > search_from:
            settlement_date_ct = to_ct(settlement_date)
            quarter_start_ct_month = int((settlement_date_ct.month - 1) / 3) * 3 + 1
            quarter_start = to_utc(
                ct_datetime(settlement_date_ct.year, quarter_start_ct_month, 1)
            )
            try:
                quarter = quarters[quarter_start]
            except KeyError:
                quarter = quarters[quarter_start] = {}

            try:
                day = quarter[settlement_date]
            except KeyError:
                day = quarter[settlement_date] = {}

            settlement_run_type = record["Settlement_Run_Type"]
            day[settlement_run_type] = record

    # Only return complate quarters
    if len(quarters) > 0:
        last_start = sorted(quarters.keys())[-1]
        del quarters[last_start]
    return quarters


def import_reconciled_daily_levy_rates(sess, log, set_progress, s):
    log("Starting to check for new LCC CfD Reconciled Daily Levy Rates")

    contract_name = "cfd_reconciled_daily_levy_rates"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    search_from = to_utc(ct_datetime(1996, 4, 1))
    for rs in sess.scalars(
        select(RateScript)
        .where(RateScript.contract == contract)
        .order_by(RateScript.start_date.desc())
    ):
        script = rs.make_script()
        try:
            records = script["records"]
        except KeyError:
            break

        complete = True
        rs_start_ct = to_ct(rs.start_date)
        quarter_finish_ct = (
            rs_start_ct + relativedelta(months=3) - relativedelta(days=1)
        )
        day_ct = rs_start_ct
        while day_ct <= quarter_finish_ct:
            try:
                records[hh_format(to_utc(day_ct))]["DF"]
            except KeyError:
                complete = False
                break

            day_ct += relativedelta(days=1)

        if complete:
            search_from = to_utc(quarter_finish_ct)
        else:
            break

    for quarter_start, quarter in _reconciled_quarters(log, s, search_from).items():

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == quarter_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, quarter_start, {"records": {}})

        gbp = 0
        kwh = 0
        run_types_reverse = list(reversed(RUN_TYPES))
        for runs in quarter.values():
            top_run = None
            for run_type in run_types_reverse:
                if run_type in runs:
                    top_run = runs[run_type]
                    break

            eligible_mwh = _parse_number(top_run["Reconciled_Eligible_Demand_MWh"])
            gbp += (
                _parse_number(top_run["Reconciled_Daily_Levy_Rate_GBP_Per_MWh"])
                * eligible_mwh
            )
            kwh += eligible_mwh * 1000

        rate = gbp / kwh

        rs_script = rs.make_script()
        records = rs_script["records"]
        for dt, runs in sorted(quarter.items()):
            date_str = hh_format(dt)
            try:
                rs_runs = records[date_str]
            except KeyError:
                rs_runs = records[date_str] = {}

            for run_type, record in runs.items():
                rs_runs[run_type] = record

        rs_script["rate_gbp_per_kwh"] = rate
        rs.update(rs_script)
        sess.commit()

    log("Finished LCC CfD Reconciled Daily Levy Rates")
    sess.commit()


def import_forecast_ilr_tra(sess, log, set_progress, s):
    log("Starting to check for new LCC CfD Forecast ILR TRA")

    contract_name = "cfd_forecast_ilr_tra"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for record in api_records(log, s, "fbece4ce-7cfc-42b7-8fb2-387cf59a3c32"):
        period_start_str = record["Period_Start"]
        if len(period_start_str) == 0:
            continue
        period_start = _parse_varying_date(period_start_str)

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == period_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, period_start, {})

        rs_script = rs.make_script()
        rs_script["record"] = record
        rs.update(rs_script)
        sess.commit()
    log("Finished LCC CfD Forecast ILR TRA")


def import_advanced_forecast_ilr_tra(sess, log, set_progress, s):
    log("Starting to check for new LCC CfD Advanced Forecast ILR TRA")

    contract_name = "cfd_advanced_forecast_ilr_tra"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for record in api_records(log, s, "e3ad6876-c1e9-46f9-b557-cb9bdae53885"):
        period_start = _parse_varying_date(record["Period Start"])

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == period_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, period_start, {})

        sensitivity = record["Sensitivity"]
        rs_script = rs.make_script()
        try:
            rs_sensitivity = rs_script["sensitivity"]
        except KeyError:
            rs_sensitivity = rs_script["sensitivity"] = {}
        rs_sensitivity[sensitivity] = record
        rs.update(rs_script)

        sess.commit()

    log("Finished LCC CfD Advanced Forecast ILR TRA")



def hh(supply_source):
    try:
        aahedc_cache = supply_source.caches["aahedc"]
    except KeyError:
        aahedc_cache = supply_source.caches["aahedc"] = {}

    for hh in supply_source.hh_data:
        hh["aahedc-kwh"] = hh["gsp-kwh"]
        try:
            rate = aahedc_cache[hh["start-date"]]
        except KeyError:
            hh_start = hh["start-date"]
            rates = supply_source.non_core_rate("aahedc", hh_start)
            rate = aahedc_cache[hh_start] = float(rates["aahedc_gbp_per_gsp_kwh"])

        hh["aahedc-rate"] = rate
        hh["aahedc-gbp"] = hh["gsp-kwh"] * rate



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import HH, ct_datetime, parse_mpan_core, to_ct, to_utc


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if not isinstance(val, Datetime):
        raise BadRequest(
            f"Problem reading {val} (of type {type(val)}) as a timestamp at "
            f"{cell.coordinate}."
        )
    return val


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


ELEM_LOOKUP = {
    "Capacity": {
        "name": "duos-availability",
        "gbp": "duos-availability-gbp",
        "units": "duos-availability-kva",
        "rate": "duos-availability-rate",
    },
    "Fixed Charge": {
        "name": "duos-fixed",
        "gbp": "duos-fixed-gbp",
        "units": "duos-fixed-days",
        "rate": "duos-fixed-rate",
    },
    "Unit Rate Zero": None,
    "Excess Reactive Power": {
        "name": "duos-reactive",
        "gbp": "duos-reactive-gbp",
        "units": "duos-reactive-kvarh",
        "rate": "duos-reactive-gbp-per-kvarh",
    },
    "Unit Rate Amber": {
        "name": "duos-amber",
        "gbp": "duos-amber-gbp",
        "units": "duos-amber-kwh",
        "rate": "duos-amber-gbp-per-kwh",
    },
    "Unit Rate Green": {
        "name": "duos-green",
        "gbp": "duos-green-gbp",
        "units": "duos-green-kwh",
        "rate": "duos-green-gbp-per-kwh",
    },
    "Unit Rate Red": {
        "name": "duos-red",
        "gbp": "duos-red-gbp",
        "units": "duos-red-kwh",
        "rate": "duos-red-gbp-per-kwh",
    },
}


def _make_raw_bills(book):
    bills = []
    sheet = book.worksheets[0]
    issue_date_raw = get_date(sheet, "H", 24)
    issue_date = to_utc(
        ct_datetime(issue_date_raw.year, issue_date_raw.month, issue_date_raw.day)
    )

    mpan_core = parse_mpan_core(str(get_int(sheet, "D", 25)))
    start_date_str = get_str(sheet, "D", 22)
    start_date_ct = to_ct(Datetime.strptime(start_date_str, "%d-%b-%Y"))
    start_date = to_utc(start_date_ct)
    finish_date_str = get_str(sheet, "F", 22)
    finish_date_ct = to_ct(Datetime.strptime(finish_date_str, "%d-%b-%Y"))
    finish_date = to_utc(finish_date_ct + relativedelta(hours=23, minutes=30))
    net = round(get_dec(sheet, "H", 33), 2)
    kwh = get_dec(sheet, "D", 33)

    breakdown = {
        "ssp-kwh": kwh,
        "ssp-gbp": net,
    }

    bills.append(
        {
            "bill_type_code": "N",
            "kwh": kwh,
            "vat": Decimal("0.00"),
            "net": net,
            "gross": net,
            "reads": [],
            "breakdown": breakdown,
            "account": mpan_core,
            "issue_date": issue_date,
            "start_date": start_date,
            "finish_date": finish_date,
            "mpan_core": mpan_core,
            "reference": "_".join(
                (
                    to_ct(start_date).strftime("%Y%m%d"),
                    to_ct(finish_date).strftime("%Y%m%d"),
                    to_ct(issue_date).strftime("%Y%m%d"),
                    mpan_core,
                    "ssp",
                )
            ),
        }
    )
    sheet = book.worksheets[3]
    for row in range(2, len(sheet["A"]) + 1):
        val = get_cell(sheet, "A", row).value
        if val is None or val == "":
            break

        start_date_str = get_str(sheet, "B", row)
        start_date_ct = to_ct(Datetime.strptime(start_date_str, "%Y-%m"))
        start_date = to_utc(start_date_ct)
        finish_date = to_utc(start_date_ct + relativedelta(months=1) - HH)

        element_desc = get_str(sheet, "D", row).strip()
        units = get_dec(sheet, "E", row)
        rate = get_dec(sheet, "F", row)
        net = round(get_dec(sheet, "G", row), 2)

        titles = ELEM_LOOKUP[element_desc]
        if titles is None:
            continue

        breakdown = {
            titles["gbp"]: net,
            titles["units"]: units,
            titles["rate"]: [rate],
        }

        bills.append(
            {
                "bill_type_code": "N",
                "kwh": Decimal(0),
                "vat": Decimal("0.00"),
                "net": net,
                "gross": net,
                "reads": [],
                "breakdown": breakdown,
                "account": mpan_core,
                "issue_date": issue_date,
                "start_date": start_date,
                "finish_date": finish_date,
                "mpan_core": mpan_core,
                "reference": "_".join(
                    (
                        to_ct(start_date).strftime("%Y%m%d"),
                        to_ct(finish_date).strftime("%Y%m%d"),
                        to_ct(issue_date).strftime("%Y%m%d"),
                        mpan_core,
                        titles["name"],
                    )
                ),
            }
        )
    return bills


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row = bills = None
        try:
            bills = _make_raw_bills(self.book)
        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills



from decimal import Decimal
from io import StringIO

from utils import match

import chellow.reports.report_429
from chellow.models import (
    BillType,
    Contract,
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    MarketRole,
    Participant,
    Site,
    User,
    insert_bill_types,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.utils import utc_datetime


def test_process_g_bill_ids(mocker):
    sess = mocker.Mock()
    forecast_date = utc_datetime(2010, 4, 1)

    query = mocker.Mock()
    sess.query.return_value = query
    m_filter = mocker.Mock()
    query.filter.return_value = m_filter
    g_bill = mocker.Mock()
    m_filter.one.return_value = g_bill
    g_bill.g_reads = []

    m_filter.order_by.return_value = []

    g_bill.g_supply = mocker.Mock()
    g_bill.start_date = forecast_date
    g_bill.finish_date = forecast_date

    MockGBill = mocker.patch("chellow.reports.report_429.GBill", autospec=True)
    MockGBill.g_supply = mocker.Mock()
    MockGBill.start_date = forecast_date
    MockGBill.finish_date = forecast_date

    find_g_era_at = g_bill.g_supply.find_g_era_at

    report_context = {}
    g_bill_ids = [1]
    bill_titles = []
    vbf = mocker.Mock()
    titles = []
    csv_writer = mocker.Mock()
    g_contract = mocker.Mock()

    chellow.reports.report_429._process_g_bill_ids(
        sess,
        report_context,
        g_contract,
        g_bill_ids,
        forecast_date,
        bill_titles,
        vbf,
        titles,
        csv_writer,
    )

    find_g_era_at.assert_not_called()


def test_batch(mocker, sess, client):
    from_date = utc_datetime(2000, 1, 1)
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    g_cv_rate_script = {
        "cvs": {
            "EA": {
                1: {"applicable_at": utc_datetime(2020, 10, 3), "cv": 39.2000},
            }
        }
    }
    GContract.insert_industry(
        sess, "cv", "", {}, utc_datetime(2000, 1, 1), None, g_cv_rate_script
    )
    ug_rate_script = {
        "ug_gbp_per_kwh": {"EA1": Decimal("40.1")},
    }
    GContract.insert_industry(
        sess, "ug", "", {}, utc_datetime(2000, 1, 1), None, ug_rate_script
    )
    ccl_rate_script = {"ccl_gbp_per_kwh": Decimal("0.00339")}
    GContract.insert_industry(sess, "ccl", "", {}, from_date, None, ccl_rate_script)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    charge_script = """
import chellow.gas.ccl
from chellow.gas.engine import g_rates
from chellow.utils import reduce_bill_hhs


def virtual_bill_titles():
    return [
        'units_consumed', 'correction_factor', 'unit_code', 'unit_factor',
        'calorific_value', 'kwh', 'gas_rate', 'gas_gbp', 'ccl_rate',
        'standing_rate', 'standing_gbp', 'net_gbp', 'vat_gbp', 'gross_gbp',
        'problem']


def virtual_bill(ds):
    chellow.gas.ccl.vb(ds)
    for hh in ds.hh_data:
        start_date = hh['start_date']
        bill_hh = ds.bill_hhs[start_date]
        bill_hh['units_consumed'] = hh['units_consumed']
        bill_hh['correction_factor'] = {hh['correction_factor']}
        bill_hh['unit_code'] = {hh['unit_code']}
        bill_hh['unit_factor'] = {hh['unit_factor']}
        bill_hh['calorific_value'] = {hh['calorific_value']}
        kwh = hh['kwh']
        bill_hh['kwh'] = kwh
        gas_rate = float(
            g_rates(ds.sess, ds.caches, db_id, start_date, False)['gas_rate'])
        bill_hh['gas_rate'] = {gas_rate}
        bill_hh['gas_gbp'] = gas_rate * kwh
        bill_hh['ccl_kwh'] = kwh
        ccl_rate = hh['ccl']
        bill_hh['ccl_rate'] = {ccl_rate}
        bill_hh['ccl_kwh'] = kwh
        bill_hh['ccl_gbp'] = kwh * ccl_rate
        if hh['utc_is_month_end']:
            standing_rate = float(
                g_rates(
                    ds.sess, ds.caches, db_id, start_date, False)['standing_rate'])
            bill_hh['standing_rate'] = {standing_rate}
            bill_hh['standing_gbp'] = standing_rate
        if hh['utc_decimal_hour'] == 0:
            pass

        bill_hh['net_gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat_gbp'] = 0
        bill_hh['gross_gbp'] = bill_hh['net_gbp'] + bill_hh['vat_gbp']

    ds.bill = reduce_bill_hhs(ds.bill_hhs)
"""
    g_contract_rate_script = {
        "gas_rate": 0.1,
        "standing_rate": 0.1,
    }
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        g_contract_rate_script,
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_n = BillType.get_by_code(sess, "N")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_429.open_file", return_value=mock_file)

    chellow.reports.report_429.content(g_batch.id, g_bill.id, None, None, None, user.id)

    actual = mock_file.getvalue()
    expected = [
        "batch,bill_reference,bill_type,bill_start_date,bill_finish_date,"
        "mprn,supply_name,site_code,site_name,covered_start,covered_finish,"
        "covered_bill_ids,covered_units_consumed,virtual_units_consumed,"
        "covered_correction_factor,virtual_correction_factor,"
        "covered_unit_code,virtual_unit_code,covered_unit_factor,"
        "virtual_unit_factor,covered_calorific_value,virtual_calorific_value,"
        "covered_kwh,virtual_kwh,covered_gas_rate,virtual_gas_rate,"
        "covered_gas_gbp,virtual_gas_gbp,difference_gas_gbp,covered_ccl_rate,"
        "virtual_ccl_rate,covered_standing_rate,virtual_standing_rate,"
        "covered_standing_gbp,virtual_standing_gbp,difference_standing_gbp,"
        "covered_net_gbp,virtual_net_gbp,difference_net_gbp,covered_vat_gbp,"
        "virtual_vat_gbp,difference_vat_gbp,covered_gross_gbp,"
        "virtual_gross_gbp,difference_gross_gbp,covered_problem,"
        "virtual_problem",
        "b1,55h883,N,2020-01-01 00:00,2020-01-31 23:30,87614362,main,22488,"
        "Water Works,2020-01-01 00:00,2020-01-31 23:30,1,771,0,,1.0,,M3,,1.0"
        ",,39.2,45,0.0,,0.1,,0.0,,,0.00339,,0.1,,0.1,,12.40,0.1,12.3,1.20,0,"
        "1.2,14.52,0.1,14.42,,",
    ]
    expected_str = "\r\n".join(expected) + "\r\n"

    assert actual == expected_str


def test_batch_http(mocker, sess, client):
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    user = User.get_by_email_address(sess, "admin@example.com")
    sess.commit()

    query_string = {"g_batch_id": g_batch.id}

    mock_Thread = mocker.patch(
        "chellow.reports.report_429.threading.Thread", autospec=True
    )
    response = client.get("/reports/429", query_string=query_string)

    match(response, 303)

    args = g_batch.id, None, None, None, None, user.id
    mock_Thread.assert_called_with(target=chellow.reports.report_429.content, args=args)


def test_bill_http(mocker, sess, client):
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_n = BillType.get_by_code(sess, "N")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    sess.commit()

    query_string = {"g_bill_id": g_bill.id}

    mock_Thread = mocker.patch(
        "chellow.reports.report_429.threading.Thread", autospec=True
    )
    response = client.get("/reports/429", query_string=query_string)

    match(response, 303)

    args = None, g_bill.id, None, None, None, user.id
    mock_Thread.assert_called_with(target=chellow.reports.report_429.content, args=args)



from chellow.e.bsuos import _process_url


def test_process_url_csv(mocker, sess):
    res = mocker.Mock()
    res.text = "Date"
    res.status_code = 200
    res.reason = "okay"
    logger = mocker.Mock()
    contract = mocker.Mock()
    url = mocker.Mock()
    s = mocker.Mock()
    s.get = mocker.Mock(return_value=res)
    _process_url(logger, sess, url, contract, s)



import csv
import threading
import traceback
from datetime import datetime as Datetime, timedelta as Timedelta

from flask import g

import pytz

from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from chellow.dloads import open_file
from chellow.models import Session, Site, Snag, User
from chellow.utils import hh_format
from chellow.views import chellow_redirect


def content(user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("site_snags.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(
                (
                    "Chellow Id",
                    "Site Code",
                    "Site Name",
                    "Snag Description",
                    "Start Date",
                    "Finish Date",
                    "Days Since Snag Finished",
                    "Duration Of Snag (Days)",
                    "Is Ignored?",
                )
            )

            now = Datetime.now(pytz.utc)

            for snag in (
                sess.query(Snag)
                .join(Site)
                .filter(Snag.site != null())
                .order_by(Site.code, Snag.description, Snag.start_date, Snag.id)
                .options(joinedload(Snag.site))
            ):
                snag_start = snag.start_date
                snag_finish = snag.finish_date
                if snag_finish is None:
                    duration = now - snag_start
                    age_of_snag = Timedelta(0)
                else:
                    duration = snag_finish - snag_start
                    age_of_snag = now - snag_finish

                writer.writerow(
                    (
                        str(snag.id),
                        snag.site.code,
                        snag.site.name,
                        snag.description,
                        hh_format(snag_start),
                        hh_format(snag_finish),
                        str(age_of_snag.days + age_of_snag.seconds / (3600 * 24)),
                        str(duration.days + duration.seconds / (3600 * 24)),
                        str(snag.is_ignored),
                    )
                )
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    args = (g.user.id,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
from io import StringIO

from chellow.edi_lib import EdiParser, to_date, to_decimal
from chellow.utils import hh_after, parse_mpan_core

read_type_map = {"00": "N", "01": "E", "02": "E", "04": "C", "06": "X", "07": "N"}


class Parser:
    def __init__(self, f):
        self.parser = EdiParser(StringIO(str(f.read(), "utf-8", errors="ignore")))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        for self.line_number, code in enumerate(self.parser):
            if code == "CLO":
                cloc = self.parser.elements[0]
                account = cloc[1]

            elif code == "BCD":
                ivdt = self.parser.elements[0]
                invn = self.parser.elements[2]
                btcd = self.parser.elements[5]

                reference = invn[0]
                bill_type_code = btcd[0]
                issue_date = to_date(ivdt[0])

            elif code == "MHD":
                typ = self.parser.elements[1]
                message_type = typ[0]
                if message_type == "UTLBIL":
                    issue_date = None
                    start_date = None
                    finish_date = None
                    account = None
                    reference = None
                    net = Decimal(0.00)
                    vat = Decimal(0.00)
                    reads = []
                    mpan_core = None

            elif code == "CCD":
                ccde = self.parser.elements[1]
                consumption_charge_indicator = ccde[0]
                charge_type = ccde[2]
                if consumption_charge_indicator != "5" and charge_type in [
                    "7",
                    "8",
                    "9",
                ]:
                    prev_read_date = to_date(self.parser.elements[7][0])
                if hh_after(start_date, prev_read_date):
                    start_date = prev_read_date
                register_finish_date = to_date(self.parser.elements[6][0])
                if finish_date is None or finish_date < register_finish_date:
                    finish_date = register_finish_date
                if charge_type == "7":
                    tmod = self.parser.elements[3]
                    mtnr = self.parser.elements[4]
                    mloc = self.parser.elements[5]
                    prrd = self.parser.elements[9]
                    adjf = self.parser.elements[12]
                    pres_read_type = read_type_map[prrd[1]]
                    prev_read_type = read_type_map[prrd[3]]
                    coefficient = Decimal(adjf[1]) / Decimal(100000)
                    pres_read_value = Decimal(prrd[0]) / Decimal(1000)
                    prev_read_value = Decimal(prrd[2]) / Decimal(1000)
                    msn = mtnr[0]
                    tpr_code = tmod[0].zfill(5)
                    read = {
                        "msn": msn,
                        "mpan": mloc[0],
                        "coefficient": coefficient,
                        "units": "kWh",
                        "tpr_code": tpr_code,
                        "prev_date": prev_read_date,
                        "prev_value": prev_read_value,
                        "prev_type_code": prev_read_type,
                        "pres_date": register_finish_date,
                        "pres_value": pres_read_value,
                        "pres_type_code": pres_read_type,
                    }
                    reads.append(read)

            elif code == "MTR":
                if message_type == "UTLBIL":
                    raw_bill = {
                        "bill_type_code": bill_type_code,
                        "account": account,
                        "mpan_core": mpan_core,
                        "reference": reference,
                        "issue_date": issue_date,
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "kwh": Decimal(0),
                        "net": net,
                        "vat": vat,
                        "gross": Decimal("0.00"),
                        "breakdown": {},
                        "reads": reads,
                    }
                    raw_bills.append(raw_bill)

            elif code == "MAN":
                madn = self.parser.elements[2]
                # pc_code = "0" + madn[3]
                # mtc_code = madn[4]
                # llfc_code = madn[5]

                mpan_core = parse_mpan_core(madn[0] + " " + madn[1] + madn[2])

            elif code == "VAT":
                uvla = self.parser.elements[5]
                net = Decimal("0.00") + to_decimal(uvla)
                uvtt = self.parser.elements[6]
                vat = Decimal("0.00") + to_decimal(uvtt)

        return raw_bills



import threading
import traceback
from datetime import datetime as Datetime

from flask import g, request

from sqlalchemy import null, or_, select, text, true

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.models import Era, Session, Site, SiteEra, Supply, User
from chellow.utils import (
    hh_format,
    hh_range,
    parse_mpan_core,
    req_bool,
    req_date,
    req_int,
    req_str,
    to_ct,
)
from chellow.views import chellow_redirect


def csv_str(row):
    frow = []
    for cell in row:
        if cell is None:
            val = ""
        elif isinstance(cell, Datetime):
            val = hh_format(cell)
        else:
            val = str(cell)
        frow.append(val)

    return ",".join('"' + v + '"' for v in frow) + "\n"


def content(start_date, finish_date, supply_id, mpan_cores, is_zipped, user_id):
    file_extension = ".zip" if is_zipped else ".csv"

    base_name = (
        f'hh_data_row_{to_ct(start_date).strftime("%Y%m%d%H%M")}{file_extension}'
    )

    tls = ["Site Code", "Imp MPAN Core", "Exp Mpan Core", "HH Start Clock-Time"]
    for polarity in ("Import", "Export"):
        for suffix in (
            "ACTIVE kWh",
            "ACTIVE Status",
            "ACTIVE Modified",
            "REACTIVE_IMP kVArh",
            "REACTIVE_IMP Status",
            "REACTIVE_IMP Modified",
            "REACTIVE_EXP kVArh",
            "REACTIVE_EXP Status",
            "REACTIVE_EXP Modified",
        ):
            tls.append(polarity + " " + suffix)

    titles = csv_str(tls)
    tmp_file = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)

            if is_zipped:
                zf = open_file(base_name, user, mode="w", is_zip=True)
            else:
                tmp_file = open_file(base_name, user, mode="w")

            caches = {}
            supplies = (
                select(Supply)
                .join(Era)
                .where(
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
                .order_by(Supply.id)
                .distinct()
            )
            if supply_id is not None:
                sup = Supply.get_by_id(sess, supply_id)
                supplies = supplies.where(Era.supply == sup)

            if mpan_cores is not None:
                supplies = supplies.where(
                    or_(
                        Era.imp_mpan_core.in_(mpan_cores),
                        Era.exp_mpan_core.in_(mpan_cores),
                    )
                )

            if not is_zipped:
                tmp_file.write(titles)

            for supply in sess.scalars(supplies):
                site, era = (
                    sess.query(Site, Era)
                    .join(Era.site_eras)
                    .filter(
                        Era.supply == supply,
                        Era.start_date <= finish_date,
                        SiteEra.site_id == Site.id,
                        or_(Era.finish_date == null(), Era.finish_date >= start_date),
                        SiteEra.is_physical == true(),
                    )
                    .order_by(Era.id)
                    .first()
                )

                outs = []
                data = iter(
                    sess.execute(
                        text(
                            """
    select hh_base.start_date,
        max(imp_active.value), max(imp_active.status),
        max(imp_active.last_modified),
        max(imp_reactive_imp.value), max(imp_reactive_imp.status),
        max(imp_reactive_imp.last_modified),
        max(imp_reactive_exp.value), max(imp_reactive_exp.status),
        max(imp_reactive_exp.last_modified),
        max(exp_active.value), max(exp_active.status),
        max(exp_active.last_modified),
        max(exp_reactive_imp.value), max(imp_reactive_imp.status),
        max(exp_reactive_imp.last_modified),
        max(exp_reactive_exp.value), max(exp_reactive_exp.status),
        max(exp_reactive_exp.last_modified)
    from hh_datum hh_base
        join channel on hh_base.channel_id = channel.id
        join era on channel.era_id = era.id
        left join hh_datum imp_active
            on (imp_active.id = hh_base.id and channel.imp_related is true and
                channel.channel_type = 'ACTIVE')
        left join hh_datum imp_reactive_imp
            on (imp_reactive_imp.id = hh_base.id
                and channel.imp_related is true and
                channel.channel_type = 'REACTIVE_IMP')
        left join hh_datum imp_reactive_exp
            on (imp_reactive_exp.id = hh_base.id
                and channel.imp_related is true and
                channel.channel_type = 'REACTIVE_EXP')
        left join hh_datum exp_active
            on (exp_active.id = hh_base.id and channel.imp_related is false and
                channel.channel_type = 'ACTIVE')
        left join hh_datum exp_reactive_imp
                    on (exp_reactive_imp.id = hh_base.id
                        and channel.imp_related is
                    false and channel.channel_type = 'REACTIVE_IMP')
        left join hh_datum exp_reactive_exp
                    on (exp_reactive_exp.id = hh_base.id
                    and channel.imp_related is false
                    and channel.channel_type = 'REACTIVE_EXP')
    where supply_id = :supply_id
        and hh_base.start_date between :start_date and :finish_date
    group by hh_base.start_date
    order by hh_base.start_date
        """
                        ),
                        params={
                            "supply_id": supply.id,
                            "start_date": start_date,
                            "finish_date": finish_date,
                        },
                    )
                )
                datum = next(data, None)

                for dt in hh_range(caches, start_date, finish_date):
                    row = [site.code, era.imp_mpan_core, era.exp_mpan_core, dt]
                    if datum is not None:
                        (
                            hh_start_date,
                            imp_active,
                            imp_active_status,
                            imp_active_modified,
                            imp_reactive_imp,
                            imp_reactive_imp_status,
                            imp_reactive_imp_modified,
                            imp_reactive_exp,
                            imp_reactive_exp_status,
                            imp_reactive_exp_modified,
                            exp_active,
                            exp_active_status,
                            exp_active_modified,
                            exp_reactive_imp,
                            exp_reactive_imp_status,
                            exp_reactive_imp_modified,
                            exp_reactive_exp,
                            exp_reactive_exp_status,
                            exp_reactive_exp_modified,
                        ) = datum
                        if hh_start_date == dt:
                            datum = next(data, None)
                            row += [
                                imp_active,
                                imp_active_status,
                                imp_active_modified,
                                imp_reactive_imp,
                                imp_reactive_imp_status,
                                imp_reactive_imp_modified,
                                imp_reactive_exp,
                                imp_reactive_exp_status,
                                imp_reactive_exp_modified,
                                exp_active,
                                exp_active_status,
                                exp_active_modified,
                                exp_reactive_imp,
                                exp_reactive_imp_status,
                                exp_reactive_imp_modified,
                                exp_reactive_exp,
                                exp_reactive_exp_status,
                                exp_reactive_exp_modified,
                            ]

                    outs.append(csv_str(row))

                if is_zipped:
                    fnm = (
                        f"hh_data_row_{era.id}_{era.imp_mpan_core}_{era.exp_mpan_core}"
                    )
                    zf.writestr(fnm.replace(" ", "") + ".csv", titles + "".join(outs))
                else:
                    tmp_file.write("".join(outs))

                # Avoid a long-running transaction
                sess.rollback()
    except BaseException:
        msg = f"Problem {traceback.format_exc()}"
        print(msg)
        if is_zipped:
            zf.writestr("error.txt", msg)
        else:
            if tmp_file is not None:
                tmp_file.write(msg)
    finally:
        if is_zipped:
            zf.close()
        else:
            if tmp_file is not None:
                tmp_file.close()


def do_post(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None

    if "mpan_cores" in request.values:
        mpan_cores_str = req_str("mpan_cores")
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    else:
        mpan_cores = None

    if finish_date < start_date:
        raise BadRequest("The finish date can't be before the start date.")

    is_zipped = req_bool("is_zipped")
    args = start_date, finish_date, supply_id, mpan_cores, is_zipped, g.user.id

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)
